import numba
import numpy as np

from ta_cn.utils import pd_to_np


@numba.jit(nopython=True, cache=True, nogil=True)
def ewm_mean_1d_nb(a, out, alpha, minp: int = 0, adjust: bool = False):
    """Return exponential weighted average.
    Numba equivalent to `pd.Series(a).ewm(span=span, min_periods=minp, adjust=adjust).mean()`.
    Adaptation of `pd._libs.window.aggregations.window_aggregations.ewma` with default arguments."""
    N = len(a)
    old_wt_factor = 1. - alpha
    new_wt_factor = alpha
    weighted_avg = a[0]
    is_observation = (weighted_avg == weighted_avg)
    nobs = int(is_observation)
    out[0] = weighted_avg if (nobs >= minp) else np.nan
    old_wt = 1.

    for i in range(1, N):
        cur = a[i]
        is_observation = (cur == cur)  # 非NaN
        nobs += is_observation

        if weighted_avg == weighted_avg:
            old_wt *= old_wt_factor[i]
            new_wt = new_wt_factor[i]
            if is_observation:
                # avoid numerical errors on constant series
                if weighted_avg != cur:
                    weighted_avg = ((old_wt * weighted_avg) + (new_wt * cur)) / (old_wt + new_wt)
                if adjust:
                    old_wt += new_wt
                else:
                    old_wt = 1.
        elif is_observation:
            weighted_avg = cur
        out[i] = weighted_avg if (nobs >= minp) else np.nan
    return out


@numba.jit(nopython=True, cache=True, nogil=True)
def ewm_mean_nb(a, out, alpha, minp: int = 0, adjust: bool = False):
    """2-dim version of `ewm_mean_1d_nb`."""
    for col in range(a.shape[1]):
        out[:, col] = ewm_mean_1d_nb(a[:, col], out[:, col], alpha[:, col], minp=minp, adjust=adjust)
    return out


def ewm_mean(a, alpha=None, span=None, com=None, min_periods: int = 0, adjust: bool = False):
    """指数移动平均

    pandas使用cython技术，本处使用numba。实测发现numba版更快

    Parameters
    ----------
    a
    alpha
    span
    com
    min_periods
    adjust:
        是否使用调整算法。一般使用False

    References
    ----------
    https://github.com/pandas-dev/pandas/blob/main/pandas/_libs/window/aggregations.pyx
    https://github.com/polakowo/vectorbt/blob/master/vectorbt/generic/nb.py

    """
    # 三个参数只使用其中一个即可
    if span is not None:
        com = (span - 1) / 2.0
    if com is not None:
        alpha = 1. / (1. + com)

    a = pd_to_np(a)
    out = np.empty_like(a)

    if isinstance(alpha, (int, float)):
        # 单数字就扩展
        alpha = np.full_like(a, fill_value=alpha)

    if a.ndim == 2:
        return ewm_mean_nb(a, out, alpha, minp=min_periods, adjust=adjust)
    else:
        return ewm_mean_1d_nb(a, out, alpha, minp=min_periods, adjust=adjust)



"""alpha101的实现

不做优化，直接复制。主要是防止在别的地方的代码优化时搞错，
在这里不做优化，只做转换，所以出错的概率小

1. ?:三元表达式要改成 IF()
2. ||改成|
3. ^改成**
4 <改成LessThan
"""
from ta_cn.imports.wq_long import *


def alpha_001(close, returns, **kwargs):
    """Alpha#1: (rank(Ts_ArgMax(SignedPower(((returns < 0) ? stddev(returns, 20) : close), 2.), 5)) - 0.5)"""
    return (rank(Ts_ArgMax(SignedPower(IF((returns < 0), stddev(returns, 20), close), 2.), 5)) - 0.5)


def alpha_002(open, close, volume, **kwargs):
    """Alpha#2: (-1 * correlation(rank(delta(log(volume), 2)), rank(((close - open) / open)), 6))"""
    return (-1 * correlation(rank(delta(log(volume), 2)), rank(((close - open) / open)), 6))


def alpha_003(open, volume, **kwargs):
    """Alpha#3: (-1 * correlation(rank(open), rank(volume), 10))"""
    return (-1 * correlation(rank(open), rank(volume), 10))


def alpha_004(low, **kwargs):
    """Alpha#4: (-1 * Ts_Rank(rank(low), 9))"""
    return (-1 * Ts_Rank(rank(low), 9))


def alpha_005(open, close, vwap, **kwargs):
    """Alpha#5: (rank((open - (sum(vwap, 10) / 10))) * (-1 * abs(rank((close - vwap)))))"""
    return (rank((open - (sum(vwap, 10) / 10))) * (-1 * abs(rank((close - vwap)))))


def alpha_006(open, volume, **kwargs):
    """Alpha#6: (-1 * correlation(open, volume, 10))"""
    return (-1 * correlation(open, volume, 10))


def alpha_007(close, volume, adv20, **kwargs):
    """Alpha#7: ((adv20 < volume) ? ((-1 * ts_rank(abs(delta(close, 7)), 60)) * sign(delta(close, 7))) : (-1* 1))"""
    return IF((adv20 < volume), ((-1 * ts_rank(abs(delta(close, 7)), 60)) * sign(delta(close, 7))), (-1 * 1))


def alpha_008(open, returns, **kwargs):
    """Alpha#8: (-1 * rank(((sum(open, 5) * sum(returns, 5)) - delay((sum(open, 5) * sum(returns, 5)), 10))))"""
    return (-1 * rank(((sum(open, 5) * sum(returns, 5)) - delay((sum(open, 5) * sum(returns, 5)), 10))))


def alpha_009(close, **kwargs):
    """Alpha#9: ((0 < ts_min(delta(close, 1), 5)) ? delta(close, 1) : ((ts_max(delta(close, 1), 5) < 0) ? delta(close, 1) : (-1 * delta(close, 1))))"""
    return IF((0 < ts_min(delta(close, 1), 5)), delta(close, 1),
              IF((ts_max(delta(close, 1), 5) < 0), delta(close, 1), (-1 * delta(close, 1))))


def alpha_010(close, **kwargs):
    """Alpha#10: rank(((0 < ts_min(delta(close, 1), 4)) ? delta(close, 1) : ((ts_max(delta(close, 1), 4) < 0) ? delta(close, 1) : (-1 * delta(close, 1)))))"""
    return rank(IF((0 < ts_min(delta(close, 1), 4)), delta(close, 1),
                   IF((ts_max(delta(close, 1), 4) < 0), delta(close, 1), (-1 * delta(close, 1)))))


def alpha_011(close, volume, vwap, **kwargs):
    """Alpha#11: ((rank(ts_max((vwap - close), 3)) + rank(ts_min((vwap - close), 3))) * rank(delta(volume, 3)))"""
    return ((rank(ts_max((vwap - close), 3)) + rank(ts_min((vwap - close), 3))) * rank(delta(volume, 3)))


def alpha_012(close, volume, **kwargs):
    """Alpha#12: (sign(delta(volume, 1)) * (-1 * delta(close, 1)))"""
    return (sign(delta(volume, 1)) * (-1 * delta(close, 1)))


def alpha_013(close, volume, **kwargs):
    """Alpha#13: (-1 * rank(covariance(rank(close), rank(volume), 5)))"""
    return (-1 * rank(covariance(rank(close), rank(volume), 5)))


def alpha_014(open, volume, returns, **kwargs):
    """Alpha#14: ((-1 * rank(delta(returns, 3))) * correlation(open, volume, 10))"""
    return ((-1 * rank(delta(returns, 3))) * correlation(open, volume, 10))


def alpha_015(high, volume, **kwargs):
    """Alpha#15: (-1 * sum(rank(correlation(rank(high), rank(volume), 3)), 3))"""
    return (-1 * sum(rank(correlation(rank(high), rank(volume), 3)), 3))


def alpha_016(high, volume, **kwargs):
    """Alpha#16: (-1 * rank(covariance(rank(high), rank(volume), 5)))"""
    return (-1 * rank(covariance(rank(high), rank(volume), 5)))


def alpha_017(close, volume, adv20, **kwargs):
    """Alpha#17: (((-1 * rank(ts_rank(close, 10))) * rank(delta(delta(close, 1), 1))) * rank(ts_rank((volume / adv20), 5)))"""
    return (((-1 * rank(ts_rank(close, 10))) * rank(delta(delta(close, 1), 1))) * rank(ts_rank((volume / adv20), 5)))


def alpha_018(open, close, **kwargs):
    """Alpha#18: (-1 * rank(((stddev(abs((close - open)), 5) + (close - open)) + correlation(close, open, 10))))"""
    return (-1 * rank(((stddev(abs((close - open)), 5) + (close - open)) + correlation(close, open, 10))))


def alpha_019(close, returns, **kwargs):
    """Alpha#19: ((-1 * sign(((close - delay(close, 7)) + delta(close, 7)))) * (1 + rank((1 + sum(returns, 250)))))"""
    return ((-1 * sign(((close - delay(close, 7)) + delta(close, 7)))) * (1 + rank((1 + sum(returns, 250)))))


def alpha_020(open, high, low, close, **kwargs):
    """Alpha#20: (((-1 * rank((open - delay(high, 1)))) * rank((open - delay(close, 1)))) * rank((open - delay(low, 1))))"""
    return (((-1 * rank((open - delay(high, 1)))) * rank((open - delay(close, 1)))) * rank((open - delay(low, 1))))


def alpha_021(close, volume, adv20, **kwargs):
    """Alpha#21: ((((sum(close, 8) / 8) + stddev(close, 8)) < (sum(close, 2) / 2)) ? (-1 * 1) : (((sum(close, 2) / 2) < ((sum(close, 8) / 8) - stddev(close, 8))) ? 1 : (((1 < (volume / adv20)) || ((volume / adv20) == 1)) ? 1 : (-1 * 1))))"""
    return IF((((sum(close, 8) / 8) + stddev(close, 8)) < (sum(close, 2) / 2)), (-1 * 1),
              IF(((sum(close, 2) / 2) < ((sum(close, 8) / 8) - stddev(close, 8))), 1,
                 IF(((1 < (volume / adv20)) | ((volume / adv20) == 1)), 1, (-1 * 1))))


def alpha_022(high, close, volume, **kwargs):
    """Alpha#22: (-1 * (delta(correlation(high, volume, 5), 5) * rank(stddev(close, 20))))"""
    return (-1 * (delta(correlation(high, volume, 5), 5) * rank(stddev(close, 20))))


def alpha_023(high, **kwargs):
    """Alpha#23: (((sum(high, 20) / 20) < high) ? (-1 * delta(high, 2)) : 0)"""
    return IF(((sum(high, 20) / 20) < high), (-1 * delta(high, 2)), 0)


def alpha_024(close, **kwargs):
    """Alpha#24: ((((delta((sum(close, 100) / 100), 100) / delay(close, 100)) < 0.05) || ((delta((sum(close, 100) / 100), 100) / delay(close, 100)) == 0.05)) ? (-1 * (close - ts_min(close, 100))) : (-1 * delta(close, 3)))"""
    return IF((((delta((sum(close, 100) / 100), 100) / delay(close, 100)) < 0.05) | (
            (delta((sum(close, 100) / 100), 100) / delay(close, 100)) == 0.05)),
              (-1 * (close - ts_min(close, 100))), (-1 * delta(close, 3)))


def alpha_025(high, close, returns, adv20, vwap, **kwargs):
    """Alpha#25: rank(((((-1 * returns) * adv20) * vwap) * (high - close)))"""
    return rank(((((-1 * returns) * adv20) * vwap) * (high - close)))


def alpha_026(high, volume, **kwargs):
    """Alpha#26: (-1 * ts_max(correlation(ts_rank(volume, 5), ts_rank(high, 5), 5), 3))"""
    return (-1 * ts_max(correlation(ts_rank(volume, 5), ts_rank(high, 5), 5), 3))


def alpha_027(volume, vwap, **kwargs):
    """Alpha#27: ((0.5 < rank((sum(correlation(rank(volume), rank(vwap), 6), 2) / 2.0))) ? (-1 * 1) : 1)"""
    return IF((0.5 < rank((sum(correlation(rank(volume), rank(vwap), 6), 2) / 2.0))), (-1 * 1), 1)


def alpha_028(high, low, close, adv20, **kwargs):
    """Alpha#28: scale(((correlation(adv20, low, 5) + ((high + low) / 2)) - close))"""
    return scale(((correlation(adv20, low, 5) + ((high + low) / 2)) - close), 1)


def alpha_029(close, returns, **kwargs):
    """Alpha#29: (min(product(rank(rank(scale(log(sum(ts_min(rank(rank((-1 * rank(delta((close - 1), 5))))), 2), 1))))), 1), 5) + ts_rank(delay((-1 * returns), 6), 5))"""
    return (min(
        product(rank(rank(scale(log(sum(ts_min(rank(rank((-1 * rank(delta((close - 1), 5))))), 2), 1)), 1))), 1),
        5) + ts_rank(delay((-1 * returns), 6), 5))


def alpha_030(close, volume, **kwargs):
    """Alpha#30: (((1.0 - rank(((sign((close - delay(close, 1))) + sign((delay(close, 1) - delay(close, 2)))) +
sign((delay(close, 2) - delay(close, 3)))))) * sum(volume, 5)) / sum(volume, 20))"""
    return (((1.0 - rank(((sign((close - delay(close, 1))) + sign((delay(close, 1) - delay(close, 2)))) +
                          sign((delay(close, 2) - delay(close, 3)))))) * sum(volume, 5)) / sum(volume, 20))


def alpha_031(low, close, adv20, **kwargs):
    """Alpha#31: ((rank(rank(rank(decay_linear((-1 * rank(rank(delta(close, 10)))), 10)))) + rank((-1 *
delta(close, 3)))) + sign(scale(correlation(adv20, low, 12))))"""
    return ((rank(rank(rank(decay_linear((-1 * rank(rank(delta(close, 10)))), 10)))) + rank((-1 *
                                                                                             delta(close, 3)))) + sign(
        scale(correlation(adv20, low, 12), 1)))


def alpha_032(close, vwap, **kwargs):
    """Alpha#32: (scale(((sum(close, 7) / 7) - close)) + (20 * scale(correlation(vwap, delay(close, 5), 230))))"""
    return (scale(((sum(close, 7) / 7) - close), 1) + (20 * scale(correlation(vwap, delay(close, 5), 230), 1)))


def alpha_033(open, close, **kwargs):
    """Alpha#33: rank((-1 * ((1 - (open / close))^1)))"""
    return rank((-1 * ((1 - (open / close)) ** 1)))


def alpha_034(close, returns, **kwargs):
    """Alpha#34: rank(((1 - rank((stddev(returns, 2) / stddev(returns, 5)))) + (1 - rank(delta(close, 1)))))"""
    return rank(((1 - rank((stddev(returns, 2) / stddev(returns, 5)))) + (1 - rank(delta(close, 1)))))


def alpha_035(high, low, close, volume, returns, **kwargs):
    """Alpha#35: ((Ts_Rank(volume, 32) * (1 - Ts_Rank(((close + high) - low), 16))) * (1 - Ts_Rank(returns, 32)))"""
    return ((Ts_Rank(volume, 32) * (1 - Ts_Rank(((close + high) - low), 16))) * (1 - Ts_Rank(returns, 32)))


def alpha_036(open, close, volume, returns, vwap, adv20, **kwargs):
    """Alpha#36: (((((2.21 * rank(correlation((close - open), delay(volume, 1), 15))) + (0.7 * rank((open
- close)))) + (0.73 * rank(Ts_Rank(delay((-1 * returns), 6), 5)))) + rank(abs(correlation(vwap,
adv20, 6)))) + (0.6 * rank((((sum(close, 200) / 200) - open) * (close - open)))))"""
    return (((((2.21 * rank(correlation((close - open), delay(volume, 1), 15))) + (0.7 * rank((open
                                                                                               - close)))) + (
                      0.73 * rank(Ts_Rank(delay((-1 * returns), 6), 5)))) + rank(abs(correlation(vwap,
                                                                                                 adv20, 6)))) + (
                    0.6 * rank((((sum(close, 200) / 200) - open) * (close - open)))))


def alpha_037(open, close, **kwargs):
    """Alpha#37: (rank(correlation(delay((open - close), 1), close, 200)) + rank((open - close)))"""
    return (rank(correlation(delay((open - close), 1), close, 200)) + rank((open - close)))


def alpha_038(open, close, **kwargs):
    """Alpha#38: ((-1 * rank(Ts_Rank(close, 10))) * rank((close / open)))"""
    return ((-1 * rank(Ts_Rank(close, 10))) * rank((close / open)))


def alpha_039(close, volume, returns, adv20, **kwargs):
    """Alpha#39: ((-1 * rank((delta(close, 7) * (1 - rank(decay_linear((volume / adv20), 9)))))) * (1 +
rank(sum(returns, 250))))"""
    return ((-1 * rank((delta(close, 7) * (1 - rank(decay_linear((volume / adv20), 9)))))) * (1 +
                                                                                              rank(sum(returns, 250))))


def alpha_040(high, volume, **kwargs):
    """Alpha#40: ((-1 * rank(stddev(high, 10))) * correlation(high, volume, 10))"""
    return ((-1 * rank(stddev(high, 10))) * correlation(high, volume, 10))


def alpha_041(high, low, vwap, **kwargs):
    """Alpha#41: (((high * low)^0.5) - vwap)"""
    return (((high * low) ** 0.5) - vwap)


def alpha_042(close, vwap, **kwargs):
    """Alpha#42: (rank((vwap - close)) / rank((vwap + close)))"""
    return (rank((vwap - close)) / rank((vwap + close)))


def alpha_043(close, volume, adv20, **kwargs):
    """Alpha#43: (ts_rank((volume / adv20), 20) * ts_rank((-1 * delta(close, 7)), 8))"""
    return (ts_rank((volume / adv20), 20) * ts_rank((-1 * delta(close, 7)), 8))


def alpha_044(high, volume, **kwargs):
    """Alpha#44: (-1 * correlation(high, rank(volume), 5))"""
    return (-1 * correlation(high, rank(volume), 5))


def alpha_045(close, volume, **kwargs):
    """Alpha#45: (-1 * ((rank((sum(delay(close, 5), 20) / 20)) * correlation(close, volume, 2)) *
rank(correlation(sum(close, 5), sum(close, 20), 2))))"""
    return (-1 * ((rank((sum(delay(close, 5), 20) / 20)) * correlation(close, volume, 2)) *
                  rank(correlation(sum(close, 5), sum(close, 20), 2))))


def alpha_046(close, **kwargs):
    """Alpha#46: ((0.25 < (((delay(close, 20) - delay(close, 10)) / 10) - ((delay(close, 10) - close) / 10))) ?
(-1 * 1) : (((((delay(close, 20) - delay(close, 10)) / 10) - ((delay(close, 10) - close) / 10)) < 0) ? 1 :
((-1 * 1) * (close - delay(close, 1)))))"""
    return IF((0.25 < (((delay(close, 20) - delay(close, 10)) / 10) - ((delay(close, 10) - close) / 10))),
              (-1 * 1), IF(((((delay(close, 20) - delay(close, 10)) / 10) - ((delay(close, 10) - close) / 10)) < 0), 1,
                           ((-1 * 1) * (close - delay(close, 1)))))


def alpha_047(high, close, volume, vwap, adv20, **kwargs):
    """Alpha#47: ((((rank((1 / close)) * volume) / adv20) * ((high * rank((high - close))) / (sum(high, 5) /
5))) - rank((vwap - delay(vwap, 5))))"""
    return ((((rank((1 / close)) * volume) / adv20) * ((high * rank((high - close))) / (sum(high, 5) /
                                                                                        5))) - rank(
        (vwap - delay(vwap, 5))))


def alpha_048(close, subindustry, **kwargs):
    """Alpha#48: (indneutralize(((correlation(delta(close, 1), delta(delay(close, 1), 1), 250) *
delta(close, 1)) / close), IndClass.subindustry) / sum(((delta(close, 1) / delay(close, 1))^2), 250))"""
    return (indneutralize(((correlation(delta(close, 1), delta(delay(close, 1), 1), 250) *
                            delta(close, 1)) / close), group=subindustry) / sum(
        ((delta(close, 1) / delay(close, 1)) ** 2), 250))


def alpha_049(close, **kwargs):
    """Alpha#49: (((((delay(close, 20) - delay(close, 10)) / 10) - ((delay(close, 10) - close) / 10)) < (-1 *
0.1)) ? 1 : ((-1 * 1) * (close - delay(close, 1))))"""
    return IF(((((delay(close, 20) - delay(close, 10)) / 10) - ((delay(close, 10) - close) / 10)) < (-1 *
                                                                                                     0.1)), 1,
              ((-1 * 1) * (close - delay(close, 1))))


def alpha_050(volume, vwap, **kwargs):
    """Alpha#50: (-1 * ts_max(rank(correlation(rank(volume), rank(vwap), 5)), 5))"""
    return (-1 * ts_max(rank(correlation(rank(volume), rank(vwap), 5)), 5))


def alpha_051(close, **kwargs):
    """Alpha#51: (((((delay(close, 20) - delay(close, 10)) / 10) - ((delay(close, 10) - close) / 10)) < (-1 *
0.05)) ? 1 : ((-1 * 1) * (close - delay(close, 1))))"""
    return IF(((((delay(close, 20) - delay(close, 10)) / 10) - ((delay(close, 10) - close) / 10)) < (-1 *
                                                                                                     0.05)), 1,
              ((-1 * 1) * (close - delay(close, 1))))


def alpha_052(low, volume, returns, **kwargs):
    """Alpha#52: ((((-1 * ts_min(low, 5)) + delay(ts_min(low, 5), 5)) * rank(((sum(returns, 240) -
sum(returns, 20)) / 220))) * ts_rank(volume, 5))"""
    return ((((-1 * ts_min(low, 5)) + delay(ts_min(low, 5), 5)) * rank(((sum(returns, 240) -
                                                                         sum(returns, 20)) / 220))) * ts_rank(volume,
                                                                                                              5))


def alpha_053(high, low, close, **kwargs):
    """Alpha#53: (-1 * delta((((close - low) - (high - close)) / (close - low)), 9))"""
    return (-1 * delta((((close - low) - (high - close)) / (close - low)), 9))


def alpha_054(open, high, low, close, **kwargs):
    """Alpha#54: ((-1 * ((low - close) * (open^5))) / ((low - high) * (close^5)))"""
    return ((-1 * ((low - close) * (open ** 5))) / ((low - high) * (close ** 5)))


def alpha_055(high, low, close, volume, **kwargs):
    """Alpha#55: (-1 * correlation(rank(((close - ts_min(low, 12)) / (ts_max(high, 12) - ts_min(low,
12)))), rank(volume), 6))"""
    return (-1 * correlation(rank(((close - ts_min(low, 12)) / (ts_max(high, 12) - ts_min(low,
                                                                                          12)))), rank(volume), 6))


def alpha_056(returns, cap, **kwargs):
    """Alpha#56: (0 - (1 * (rank((sum(returns, 10) / sum(sum(returns, 2), 3))) * rank((returns * cap)))))"""
    return (0 - (1 * (rank((sum(returns, 10) / sum(sum(returns, 2), 3))) * rank((returns * cap)))))


def alpha_057(close, vwap, **kwargs):
    """Alpha#57: (0 - (1 * ((close - vwap) / decay_linear(rank(ts_argmax(close, 30)), 2))))"""
    return (0 - (1 * ((close - vwap) / decay_linear(rank(ts_argmax(close, 30)), 2))))


def alpha_058(volume, vwap, sector, **kwargs):
    """Alpha#58: (-1 * Ts_Rank(decay_linear(correlation(IndNeutralize(vwap, IndClass.sector), volume, 3.92795), 7.89291), 5.50322))"""
    return (-1 * Ts_Rank(decay_linear(correlation(IndNeutralize(vwap, group=sector), volume, 3.92795), 7.89291),
                         5.50322))


def alpha_059(volume, vwap, industry, **kwargs):
    """Alpha#59: (-1 * Ts_Rank(decay_linear(correlation(IndNeutralize(((vwap * 0.728317) + (vwap * (1 - 0.728317))), IndClass.industry), volume, 4.25197), 16.2289), 8.19648))"""
    return (-1 * Ts_Rank(decay_linear(
        correlation(IndNeutralize(((vwap * 0.728317) + (vwap * (1 - 0.728317))), group=industry), volume, 4.25197),
        16.2289), 8.19648))


def alpha_060(high, low, close, volume, **kwargs):
    """Alpha#60: (0 - (1 * ((2 * scale(rank(((((close - low) - (high - close)) / (high - low)) * volume)))) -
scale(rank(ts_argmax(close, 10))))))"""
    return (0 - (1 * ((2 * scale(rank(((((close - low) - (high - close)) / (high - low)) * volume)), 1)) -
                      scale(rank(ts_argmax(close, 10)), 1))))


def alpha_061(vwap, adv180, **kwargs):
    """Alpha#61: (rank((vwap - ts_min(vwap, 16.1219))) < rank(correlation(vwap, adv180, 17.9282)))"""
    return LessThan(rank((vwap - ts_min(vwap, 16.1219))), rank(correlation(vwap, adv180, 17.9282)))


def alpha_062(open, high, low, vwap, adv20, **kwargs):
    """Alpha#62: ((rank(correlation(vwap, sum(adv20, 22.4101), 9.91009)) < rank(((rank(open) +
rank(open)) < (rank(((high + low) / 2)) + rank(high))))) * -1)"""
    return (LessThan(rank(correlation(vwap, sum(adv20, 22.4101), 9.91009)),
                     rank(((rank(open) + rank(open)) < (rank(((high + low) / 2)) + rank(high))))) * -1)


def alpha_063(open, close, vwap, adv180, industry, **kwargs):
    """Alpha#63: ((rank(decay_linear(delta(IndNeutralize(close, IndClass.industry), 2.25164), 8.22237))
- rank(decay_linear(correlation(((vwap * 0.318108) + (open * (1 - 0.318108))), sum(adv180,
37.2467), 13.557), 12.2883))) * -1)"""
    return ((rank(decay_linear(delta(IndNeutralize(close, group=industry), 2.25164), 8.22237))
             - rank(decay_linear(correlation(((vwap * 0.318108) + (open * (1 - 0.318108))), sum(adv180,
                                                                                                37.2467), 13.557),
                                 12.2883))) * -1)


def alpha_064(open, high, low, vwap, adv120, **kwargs):
    """Alpha#64: ((rank(correlation(sum(((open * 0.178404) + (low * (1 - 0.178404))), 12.7054),
sum(adv120, 12.7054), 16.6208)) < rank(delta(((((high + low) / 2) * 0.178404) + (vwap * (1 -
0.178404))), 3.69741))) * -1)"""
    return (LessThan(rank(correlation(sum(((open * 0.178404) + (low * (1 - 0.178404))), 12.7054),
                                      sum(adv120, 12.7054), 16.6208)),
                     rank(delta(((((high + low) / 2) * 0.178404) + (vwap * (1 -
                                                                            0.178404))), 3.69741))) * -1)


def alpha_065(open, vwap, adv60, **kwargs):
    """Alpha#65: ((rank(correlation(((open * 0.00817205) + (vwap * (1 - 0.00817205))), sum(adv60,
8.6911), 6.40374)) < rank((open - ts_min(open, 13.635)))) * -1)"""
    return (LessThan(rank(correlation(((open * 0.00817205) + (vwap * (1 - 0.00817205))), sum(adv60,
                                                                                             8.6911), 6.40374)),
                     rank((open - ts_min(open, 13.635)))) * -1)


def alpha_066(open, high, low, vwap, **kwargs):
    """Alpha#66: ((rank(decay_linear(delta(vwap, 3.51013), 7.23052)) + Ts_Rank(decay_linear(((((low
* 0.96633) + (low * (1 - 0.96633))) - vwap) / (open - ((high + low) / 2))), 11.4157), 6.72611)) * -1)"""
    return ((rank(decay_linear(delta(vwap, 3.51013), 7.23052)) + Ts_Rank(decay_linear(((((low
                                                                                          * 0.96633) + (low * (
            1 - 0.96633))) - vwap) / (open - ((high + low) / 2))), 11.4157), 6.72611)) * -1)


def alpha_067(high, vwap, adv20, sector, subindustry, **kwargs):
    """Alpha#67: ((rank((high - ts_min(high, 2.14593)))^rank(correlation(IndNeutralize(vwap,
IndClass.sector), IndNeutralize(adv20, IndClass.subindustry), 6.02936))) * -1)"""
    return ((rank((high - ts_min(high, 2.14593))) ** rank(correlation(IndNeutralize(vwap,
                                                                                    group=sector),
                                                                      IndNeutralize(adv20, group=subindustry),
                                                                      6.02936))) * -1)


def alpha_068(high, low, close, adv15, **kwargs):
    """Alpha#68: ((Ts_Rank(correlation(rank(high), rank(adv15), 8.91644), 13.9333) <
rank(delta(((close * 0.518371) + (low * (1 - 0.518371))), 1.06157))) * -1)"""
    return (LessThan(Ts_Rank(correlation(rank(high), rank(adv15), 8.91644), 13.9333),
                     rank(delta(((close * 0.518371) + (low * (1 - 0.518371))), 1.06157))) * -1)


def alpha_069(close, vwap, adv20, industry, **kwargs):
    """Alpha#69: ((rank(ts_max(delta(IndNeutralize(vwap, IndClass.industry), 2.72412),
4.79344))^Ts_Rank(correlation(((close * 0.490655) + (vwap * (1 - 0.490655))), adv20, 4.92416),
9.0615)) * -1)"""
    return ((rank(ts_max(delta(IndNeutralize(vwap, group=industry), 2.72412),
                         4.79344)) ** Ts_Rank(
        correlation(((close * 0.490655) + (vwap * (1 - 0.490655))), adv20, 4.92416),
        9.0615)) * -1)


def alpha_070(close, vwap, adv50, industry, **kwargs):
    """Alpha#70: ((rank(delta(vwap, 1.29456))^Ts_Rank(correlation(IndNeutralize(close,
IndClass.industry), adv50, 17.8256), 17.9171)) * -1)"""
    return ((rank(delta(vwap, 1.29456)) ** Ts_Rank(correlation(IndNeutralize(close,
                                                                             group=industry), adv50, 17.8256),
                                                   17.9171)) * -1)


def alpha_071(open, low, close, vwap, adv180, **kwargs):
    """Alpha#71: max(Ts_Rank(decay_linear(correlation(Ts_Rank(close, 3.43976), Ts_Rank(adv180,
12.0647), 18.0175), 4.20501), 15.6948), Ts_Rank(decay_linear((rank(((low + open) - (vwap +
vwap)))^2), 16.4662), 4.4388))"""
    return max(Ts_Rank(decay_linear(correlation(Ts_Rank(close, 3.43976), Ts_Rank(adv180,
                                                                                 12.0647), 18.0175), 4.20501), 15.6948),
               Ts_Rank(decay_linear((rank(((low + open) - (vwap +
                                                           vwap))) ** 2), 16.4662), 4.4388))


def alpha_072(high, low, volume, vwap, adv40, **kwargs):
    """Alpha#72: (rank(decay_linear(correlation(((high + low) / 2), adv40, 8.93345), 10.1519)) /
rank(decay_linear(correlation(Ts_Rank(vwap, 3.72469), Ts_Rank(volume, 18.5188), 6.86671),
2.95011)))"""
    return (rank(decay_linear(correlation(((high + low) / 2), adv40, 8.93345), 10.1519)) /
            rank(decay_linear(correlation(Ts_Rank(vwap, 3.72469), Ts_Rank(volume, 18.5188), 6.86671),
                              2.95011)))


def alpha_073(open, low, vwap, **kwargs):
    """Alpha#73: (max(rank(decay_linear(delta(vwap, 4.72775), 2.91864)),
Ts_Rank(decay_linear(((delta(((open * 0.147155) + (low * (1 - 0.147155))), 2.03608) / ((open *
0.147155) + (low * (1 - 0.147155)))) * -1), 3.33829), 16.7411)) * -1)"""
    return (max(rank(decay_linear(delta(vwap, 4.72775), 2.91864)),
                Ts_Rank(decay_linear(((delta(((open * 0.147155) + (low * (1 - 0.147155))), 2.03608) / ((open *
                                                                                                        0.147155) + (
                                                                                                               low * (
                                                                                                               1 - 0.147155)))) * -1),
                                     3.33829), 16.7411)) * -1)


def alpha_074(high, close, volume, vwap, adv30, **kwargs):
    """Alpha#74: ((rank(correlation(close, sum(adv30, 37.4843), 15.1365)) <
rank(correlation(rank(((high * 0.0261661) + (vwap * (1 - 0.0261661)))), rank(volume), 11.4791)))
* -1)"""
    return (LessThan(rank(correlation(close, sum(adv30, 37.4843), 15.1365)),
                     rank(correlation(rank(((high * 0.0261661) + (vwap * (1 - 0.0261661)))), rank(volume), 11.4791)))
            * -1)


def alpha_075(low, volume, vwap, adv50, **kwargs):
    """Alpha#75: (rank(correlation(vwap, volume, 4.24304)) < rank(correlation(rank(low), rank(adv50),
12.4413)))"""
    return (rank(correlation(vwap, volume, 4.24304)) < rank(correlation(rank(low), rank(adv50),
                                                                        12.4413)))


def alpha_076(low, vwap, adv81, sector, **kwargs):
    """Alpha#76: (max(rank(decay_linear(delta(vwap, 1.24383), 11.8259)),
Ts_Rank(decay_linear(Ts_Rank(correlation(IndNeutralize(low, IndClass.sector), adv81,
8.14941), 19.569), 17.1543), 19.383)) * -1)"""
    return (max(rank(decay_linear(delta(vwap, 1.24383), 11.8259)),
                Ts_Rank(decay_linear(Ts_Rank(correlation(IndNeutralize(low, group=sector), adv81,
                                                         8.14941), 19.569), 17.1543), 19.383)) * -1)


def alpha_077(high, low, vwap, adv40, **kwargs):
    """Alpha#77: min(rank(decay_linear(((((high + low) / 2) + high) - (vwap + high)), 20.0451)),
rank(decay_linear(correlation(((high + low) / 2), adv40, 3.1614), 5.64125)))"""
    return min(rank(decay_linear(((((high + low) / 2) + high) - (vwap + high)), 20.0451)),
               rank(decay_linear(correlation(((high + low) / 2), adv40, 3.1614), 5.64125)))


def alpha_078(low, volume, vwap, adv40, **kwargs):
    """Alpha#78: (rank(correlation(sum(((low * 0.352233) + (vwap * (1 - 0.352233))), 19.7428),
sum(adv40, 19.7428), 6.83313))^rank(correlation(rank(vwap), rank(volume), 5.77492)))"""
    return (rank(correlation(sum(((low * 0.352233) + (vwap * (1 - 0.352233))), 19.7428),
                             sum(adv40, 19.7428), 6.83313)) ** rank(correlation(rank(vwap), rank(volume), 5.77492)))


def alpha_079(open, close, vwap, adv150, sector, **kwargs):
    """Alpha#79: (rank(delta(IndNeutralize(((close * 0.60733) + (open * (1 - 0.60733))),
IndClass.sector), 1.23438)) < rank(correlation(Ts_Rank(vwap, 3.60973), Ts_Rank(adv150,
9.18637), 14.6644)))"""
    return LessThan(rank(delta(IndNeutralize(((close * 0.60733) + (open * (1 - 0.60733))),
                                             group=sector), 1.23438)),
                    rank(correlation(Ts_Rank(vwap, 3.60973), Ts_Rank(adv150,
                                                                     9.18637),
                                     14.6644)))


def alpha_080(open, high, adv10, industry, **kwargs):
    """Alpha#80: ((rank(Sign(delta(IndNeutralize(((open * 0.868128) + (high * (1 - 0.868128))),
IndClass.industry), 4.04545)))^Ts_Rank(correlation(high, adv10, 5.11456), 5.53756)) * -1)"""
    return ((rank(Sign(delta(IndNeutralize(((open * 0.868128) + (high * (1 - 0.868128))),
                                           group=industry), 4.04545))) ** Ts_Rank(correlation(high, adv10, 5.11456),
                                                                                  5.53756)) * -1)


def alpha_081(volume, vwap, adv10, **kwargs):
    """Alpha#81: ((rank(Log(product(rank((rank(correlation(vwap, sum(adv10, 49.6054),
8.47743))^4)), 14.9655))) < rank(correlation(rank(vwap), rank(volume), 5.07914))) * -1)"""
    return (LessThan(rank(Log(product(rank((rank(correlation(vwap, sum(adv10, 49.6054),
                                                             8.47743)) ** 4)), 14.9655))), rank(
        correlation(rank(vwap), rank(volume), 5.07914))) * -1)


def alpha_082(open, volume, sector, **kwargs):
    """Alpha#82: (min(rank(decay_linear(delta(open, 1.46063), 14.8717)),
Ts_Rank(decay_linear(correlation(IndNeutralize(volume, IndClass.sector), ((open * 0.634196) +
(open * (1 - 0.634196))), 17.4842), 6.92131), 13.4283)) * -1)"""
    return (min(rank(decay_linear(delta(open, 1.46063), 14.8717)),
                Ts_Rank(decay_linear(correlation(IndNeutralize(volume, group=sector), ((open * 0.634196) +
                                                                                       (open * (1 - 0.634196))),
                                                 17.4842), 6.92131), 13.4283)) * -1)


def alpha_083(high, low, close, volume, vwap, **kwargs):
    """Alpha#83: ((rank(delay(((high - low) / (sum(close, 5) / 5)), 2)) * rank(rank(volume))) / (((high -
low) / (sum(close, 5) / 5)) / (vwap - close)))"""
    return ((rank(delay(((high - low) / (sum(close, 5) / 5)), 2)) * rank(rank(volume))) / (((high -
                                                                                             low) / (sum(close,
                                                                                                         5) / 5)) / (
                                                                                                   vwap - close)))


def alpha_084(close, vwap, **kwargs):
    """Alpha#84: SignedPower(Ts_Rank((vwap - ts_max(vwap, 15.3217)), 20.7127), delta(close,
4.96796))"""
    return SignedPower(Ts_Rank((vwap - ts_max(vwap, 15.3217)), 20.7127), delta(close,
                                                                               4.96796))


def alpha_085(high, low, close, volume, adv30, **kwargs):
    """Alpha#85: (rank(correlation(((high * 0.876703) + (close * (1 - 0.876703))), adv30,
9.61331))^rank(correlation(Ts_Rank(((high + low) / 2), 3.70596), Ts_Rank(volume, 10.1595),
7.11408)))"""
    return (rank(correlation(((high * 0.876703) + (close * (1 - 0.876703))), adv30,
                             9.61331)) ** rank(
        correlation(Ts_Rank(((high + low) / 2), 3.70596), Ts_Rank(volume, 10.1595),
                    7.11408)))


def alpha_086(open, close, vwap, adv20, **kwargs):
    """Alpha#86: ((Ts_Rank(correlation(close, sum(adv20, 14.7444), 6.00049), 20.4195) < rank(((open
+ close) - (vwap + open)))) * -1)"""
    return (LessThan(Ts_Rank(correlation(close, sum(adv20, 14.7444), 6.00049), 20.4195), rank(((open
                                                                                                + close) - (
                                                                                                       vwap + open)))) * -1)


def alpha_087(close, vwap, adv81, industry, **kwargs):
    """Alpha#87: (max(rank(decay_linear(delta(((close * 0.369701) + (vwap * (1 - 0.369701))),
1.91233), 2.65461)), Ts_Rank(decay_linear(abs(correlation(IndNeutralize(adv81,
IndClass.industry), close, 13.4132)), 4.89768), 14.4535)) * -1)"""
    return (max(rank(decay_linear(delta(((close * 0.369701) + (vwap * (1 - 0.369701))),
                                        1.91233), 2.65461)), Ts_Rank(decay_linear(abs(correlation(IndNeutralize(adv81,
                                                                                                                group=industry),
                                                                                                  close, 13.4132)),
                                                                                  4.89768), 14.4535)) * -1)


def alpha_088(open, high, low, close, adv60, **kwargs):
    """Alpha#88: min(rank(decay_linear(((rank(open) + rank(low)) - (rank(high) + rank(close))),
8.06882)), Ts_Rank(decay_linear(correlation(Ts_Rank(close, 8.44728), Ts_Rank(adv60,
20.6966), 8.01266), 6.65053), 2.61957))"""
    return min(rank(decay_linear(((rank(open) + rank(low)) - (rank(high) + rank(close))),
                                 8.06882)), Ts_Rank(decay_linear(correlation(Ts_Rank(close, 8.44728), Ts_Rank(adv60,
                                                                                                              20.6966),
                                                                             8.01266), 6.65053), 2.61957))


def alpha_089(low, vwap, adv10, industry, **kwargs):
    """Alpha#89: (Ts_Rank(decay_linear(correlation(((low * 0.967285) + (low * (1 - 0.967285))), adv10,
6.94279), 5.51607), 3.79744) - Ts_Rank(decay_linear(delta(IndNeutralize(vwap,
IndClass.industry), 3.48158), 10.1466), 15.3012))"""
    return (Ts_Rank(decay_linear(correlation(((low * 0.967285) + (low * (1 - 0.967285))), adv10,
                                             6.94279), 5.51607), 3.79744) - Ts_Rank(
        decay_linear(delta(IndNeutralize(vwap,
                                         group=industry), 3.48158), 10.1466), 15.3012))


def alpha_090(low, close, adv40, subindustry, **kwargs):
    """Alpha#90: ((rank((close - ts_max(close, 4.66719)))^Ts_Rank(correlation(IndNeutralize(adv40,
IndClass.subindustry), low, 5.38375), 3.21856)) * -1)"""
    return ((rank((close - ts_max(close, 4.66719))) ** Ts_Rank(correlation(IndNeutralize(adv40,
                                                                                         group=subindustry), low,
                                                                           5.38375), 3.21856)) * -1)


def alpha_091(close, volume, vwap, adv30, industry, **kwargs):
    """Alpha#91: ((Ts_Rank(decay_linear(decay_linear(correlation(IndNeutralize(close,
IndClass.industry), volume, 9.74928), 16.398), 3.83219), 4.8667) -
rank(decay_linear(correlation(vwap, adv30, 4.01303), 2.6809))) * -1)"""
    return ((Ts_Rank(decay_linear(decay_linear(correlation(IndNeutralize(close,
                                                                         group=industry), volume, 9.74928), 16.398),
                                  3.83219), 4.8667) -
             rank(decay_linear(correlation(vwap, adv30, 4.01303), 2.6809))) * -1)


def alpha_092(open, high, low, close, adv30, **kwargs):
    """Alpha#92: min(Ts_Rank(decay_linear(((((high + low) / 2) + close) < (low + open)), 14.7221),
18.8683), Ts_Rank(decay_linear(correlation(rank(low), rank(adv30), 7.58555), 6.94024),
6.80584))"""
    return min(Ts_Rank(decay_linear(((((high + low) / 2) + close) < (low + open)) * 1., 14.7221),
                       18.8683), Ts_Rank(decay_linear(correlation(rank(low), rank(adv30), 7.58555), 6.94024),
                                         6.80584))


def alpha_093(close, vwap, adv81, industry, **kwargs):
    """Alpha#93: (Ts_Rank(decay_linear(correlation(IndNeutralize(vwap, IndClass.industry), adv81,
17.4193), 19.848), 7.54455) / rank(decay_linear(delta(((close * 0.524434) + (vwap * (1 -
0.524434))), 2.77377), 16.2664)))"""
    return (Ts_Rank(decay_linear(correlation(IndNeutralize(vwap, group=industry), adv81,
                                             17.4193), 19.848), 7.54455) / rank(
        decay_linear(delta(((close * 0.524434) + (vwap * (1 -
                                                          0.524434))), 2.77377), 16.2664)))


def alpha_094(vwap, adv60, **kwargs):
    """Alpha#94: ((rank((vwap - ts_min(vwap, 11.5783)))^Ts_Rank(correlation(Ts_Rank(vwap,
19.6462), Ts_Rank(adv60, 4.02992), 18.0926), 2.70756)) * -1)"""
    return ((rank((vwap - ts_min(vwap, 11.5783))) ** Ts_Rank(correlation(Ts_Rank(vwap,
                                                                                 19.6462), Ts_Rank(adv60, 4.02992),
                                                                         18.0926), 2.70756)) * -1)


def alpha_095(open, high, low, adv40, **kwargs):
    """Alpha#95: (rank((open - ts_min(open, 12.4105))) < Ts_Rank((rank(correlation(sum(((high + low)
/ 2), 19.1351), sum(adv40, 19.1351), 12.8742))^5), 11.7584))"""
    return LessThan(rank((open - ts_min(open, 12.4105))), Ts_Rank((rank(correlation(sum(((high + low)
                                                                                         / 2), 19.1351),
                                                                                    sum(adv40, 19.1351),
                                                                                    12.8742)) ** 5), 11.7584))


def alpha_096(close, volume, vwap, adv60, **kwargs):
    """Alpha#96: (max(Ts_Rank(decay_linear(correlation(rank(vwap), rank(volume), 3.83878),
4.16783), 8.38151), Ts_Rank(decay_linear(Ts_ArgMax(correlation(Ts_Rank(close, 7.45404),
Ts_Rank(adv60, 4.13242), 3.65459), 12.6556), 14.0365), 13.4143)) * -1)"""
    return (max(Ts_Rank(decay_linear(correlation(rank(vwap), rank(volume), 3.83878),
                                     4.16783), 8.38151),
                Ts_Rank(decay_linear(Ts_ArgMax(correlation(Ts_Rank(close, 7.45404),
                                                           Ts_Rank(adv60, 4.13242), 3.65459), 12.6556), 14.0365),
                        13.4143)) * -1)


def alpha_097(low, vwap, adv60, industry, **kwargs):
    """Alpha#97: ((rank(decay_linear(delta(IndNeutralize(((low * 0.721001) + (vwap * (1 - 0.721001))),
IndClass.industry), 3.3705), 20.4523)) - Ts_Rank(decay_linear(Ts_Rank(correlation(Ts_Rank(low,
7.87871), Ts_Rank(adv60, 17.255), 4.97547), 18.5925), 15.7152), 6.71659)) * -1)"""
    return ((rank(decay_linear(delta(IndNeutralize(((low * 0.721001) + (vwap * (1 - 0.721001))),
                                                   group=industry), 3.3705), 20.4523)) - Ts_Rank(
        decay_linear(Ts_Rank(correlation(Ts_Rank(low,
                                                 7.87871), Ts_Rank(adv60, 17.255), 4.97547), 18.5925), 15.7152),
        6.71659)) * -1)


def alpha_098(open, vwap, adv5, adv15, **kwargs):
    """Alpha#98: (rank(decay_linear(correlation(vwap, sum(adv5, 26.4719), 4.58418), 7.18088)) -
rank(decay_linear(Ts_Rank(Ts_ArgMin(correlation(rank(open), rank(adv15), 20.8187), 8.62571),
6.95668), 8.07206)))"""
    return (rank(decay_linear(correlation(vwap, sum(adv5, 26.4719), 4.58418), 7.18088)) -
            rank(decay_linear(Ts_Rank(Ts_ArgMin(correlation(rank(open), rank(adv15), 20.8187), 8.62571),
                                      6.95668), 8.07206)))


def alpha_099(high, low, volume, adv60, **kwargs):
    """Alpha#99: ((rank(correlation(sum(((high + low) / 2), 19.8975), sum(adv60, 19.8975), 8.8136)) <
rank(correlation(low, volume, 6.28259))) * -1)"""
    return (LessThan(rank(correlation(sum(((high + low) / 2), 19.8975), sum(adv60, 19.8975), 8.8136)),
                     rank(correlation(low, volume, 6.28259))) * -1)


def alpha_100(high, low, close, volume, adv20, subindustry, **kwargs):
    """Alpha#100: (0 - (1 * (((1.5 * scale(indneutralize(indneutralize(rank(((((close - low) - (high -
close)) / (high - low)) * volume)), IndClass.subindustry), IndClass.subindustry))) -
scale(indneutralize((correlation(close, rank(adv20), 5) - rank(ts_argmin(close, 30))),
IndClass.subindustry))) * (volume / adv20))))"""
    return (0 - (1 * (((1.5 * scale(indneutralize(indneutralize(rank(((((close - low) - (high -
                                                                                         close)) / (
                                                                               high - low)) * volume)),
                                                                group=subindustry), group=subindustry), 1)) -
                       scale(indneutralize((correlation(close, rank(adv20), 5) - rank(ts_argmin(close, 30))),
                                           group=subindustry), 1)) * (volume / adv20))))


def alpha_101(open, high, low, close, **kwargs):
    """Alpha#101: ((close - open) / ((high - low) + .001))"""
    return ((close - open) / ((high - low) + .001))



import numba
import numpy as _np

from .utils import pd_to_np


@numba.jit(nopython=True, cache=True, nogil=True)
def _filter_nb(arr, n):
    """内部函数，请勿直接调用，请参考"""
    is_1d = arr.ndim == 1
    x = arr.shape[0]
    y = 1 if is_1d else arr.shape[1]

    for j in range(y):
        a = arr if is_1d else arr[:, j]

        # 为了跳过不必要部分，由for改while
        i = 0
        while i < x:
            if a[i]:
                a[i + 1:i + 1 + n] = 0
                i += n + 1
            else:
                i += 1
    return arr


@numba.jit(nopython=True, cache=True, nogil=True)
def _bars_last_nb(arr, out):
    """上一次条件成立到当前的周期数"""
    is_1d = arr.ndim == 1
    x = arr.shape[0]
    y = 1 if is_1d else arr.shape[1]

    for j in range(y):
        a = arr if is_1d else arr[:, j]
        b = out if is_1d else out[:, j]
        s = 0
        for i in range(x):
            if a[i]:
                s = 0
            b[i] = s
            s += 1

    return out


@numba.jit(nopython=True, cache=True, nogil=True)
def _bars_last_count_nb(arr, out):
    """

    Parameters
    ----------
    arr
    out

    References
    ----------
    https://stackoverflow.com/questions/18196811/cumsum-reset-at-nan

    """
    is_1d = arr.ndim == 1
    x = arr.shape[0]
    y = 1 if is_1d else arr.shape[1]

    for j in range(y):
        a = arr if is_1d else arr[:, j]
        b = out if is_1d else out[:, j]
        s = 0
        for i in range(x):
            if a[i]:
                s += 1
                b[i] = s
            else:
                s = 0

    return out


@numba.jit(nopython=True, cache=True, nogil=True)
def _fill_notna_nb(arr, fill_value, n: int):
    is_1d = arr.ndim == 1
    x = arr.shape[0]
    y = 1 if is_1d else arr.shape[1]

    for j in range(y):
        a = arr if is_1d else arr[:, j]
        k = n
        for i in range(x):
            cur = a[i]
            if cur != cur:
                continue
            a[i] = fill_value
            k -= 1
            if k <= 0:
                break

    return arr


@numba.jit(nopython=True, cache=True, nogil=True)
def _sum_1st_nb(arr, n):
    """前部分数据求和，之前的设置成0"""
    is_1d = arr.ndim == 1
    x = arr.shape[0]
    y = 1 if is_1d else arr.shape[1]

    for j in range(y):
        a = arr if is_1d else arr[:, j]
        k = n
        s = 0.0
        skip_nan = True
        for i in range(x):
            cur = a[i]
            if cur != cur:  # 用来判断NaN
                if skip_nan:
                    continue
                else:
                    a[i] = 0
            skip_nan = False
            s += cur
            a[i] = 0
            k -= 1
            if k <= 0:
                a[i] = s
                break
    return arr


@numba.jit(nopython=True, cache=True, nogil=True)
def _ma_1st_nb(arr, n):
    """前部分数据求和，之前的设置成0"""
    is_1d = arr.ndim == 1
    x = arr.shape[0]
    y = 1 if is_1d else arr.shape[1]

    for j in range(y):
        a = arr if is_1d else arr[:, j]
        k = n
        s = 0.0
        for i in range(x):
            cur = a[i]
            if cur != cur:  # 用来判断NaN
                continue
            s += cur
            a[i] = _np.nan
            k -= 1
            if k <= 0:
                a[i] = s / n
                break
    return arr


def fill_notna(arr, fill_value, n=1):
    """查找第一个非nan,将其改成fill_value"""
    if n < 1:
        return arr
    arr = pd_to_np(arr, copy=True)
    return _fill_notna_nb(arr, fill_value, n)


def ma_1st(arr, n=1):
    """前部分数据求平均，之前的设置成np.nan"""
    if n < 1:
        return arr
    arr = pd_to_np(arr, copy=True)
    return _ma_1st_nb(arr, n)


def sum_1st(arr, n=1):
    """前部分数据求平均，之前的设置成np.nan"""
    if n < 1:
        return arr
    arr = pd_to_np(arr, copy=True)
    return _sum_1st_nb(arr, n)


@numba.jit(nopython=True, cache=True, nogil=True)
def _avedev_nb(a):
    """avedev平均绝对偏差"""
    return _np.mean(_np.abs(a - _np.mean(a)))


@numba.jit(nopython=True, cache=True, nogil=True)
def _bars_since_n_nb(a, n):
    """BARSSINCEN(X,N):N周期内第一次X不为0到现在的天数"""
    for i, x in enumerate(a):
        if x:
            return n - i - 1
    return 0


@numba.jit(nopython=True, cache=True, nogil=True)
def _product_nb(a):
    return _np.nanprod(a)


@numba.jit(nopython=True, cache=True, nogil=True)
def _last_nb(arr, n, m):
    """LAST(X,A,B)，A>B，表示从前A日到前B日一致满足X条件"""
    return _np.all(arr[:n - m])


@numba.jit(nopython=True, cache=True, nogil=True)
def _cov_nb(arr0, arr1):
    """协方差矩阵"""
    return _np.cov(arr0, arr1)[0, 1]


@numba.jit(nopython=True, cache=True, nogil=True)
def _rolling_func_2_nb(arr0, arr1, out, timeperiod, func, *args):
    """滚动函数，在二维或三维数上遍历，二组输入

    Parameters
    ----------
    arr0:
        输入。二维或三维
    out:
        输出。一维或二维。降了一个维度
    window: int
        窗口长度
    func:
        单向量处理函数
    args:
        func的位置参数

    Returns
    -------
    np.ndarray
        一维或二维数组

    """
    if arr0.ndim == 3:
        for i, (aa, bb) in enumerate(zip(arr0, arr1)):
            for j, (a, b) in enumerate(zip(aa, bb)):
                out[i + timeperiod - 1, j] = func(a, b, *args)
    elif arr0.ndim == 2:
        for i, (a, b) in enumerate(zip(arr0, arr1)):
            out[i + timeperiod - 1] = func(a, b, *args)

    return out


@numba.jit(nopython=True, cache=True, nogil=True)
def _rolling_func_1_nb(arr, out, timeperiod, func, *args):
    """滚动函数，在二维或三维数上遍历，一组输入

    Parameters
    ----------
    arr:
        输入。二维或三维
    out:
        输出。一维或二维。降了一个维度
    window: int
        窗口长度
    func:
        单向量处理函数
    args:
        func的位置参数

    Returns
    -------
    np.ndarray
        一维或二维数组

    """
    if arr.ndim == 3:
        for i, aa in enumerate(arr):
            for j, a in enumerate(aa):
                out[i + timeperiod - 1, j] = func(a, *args)
    elif arr.ndim == 2:
        for i, a in enumerate(arr):
            out[i + timeperiod - 1] = func(a, *args)

    return out


def numpy_rolling_apply(inputs, window, func1, func2, *args):
    """滚动应用方法 处理两个"""
    out = _np.empty_like(inputs[0])
    try:
        # 可能出现类似int无法设置nan的情况
        out[:window] = _np.nan
    except:
        out[:window] = 0

    arrs = [_np.lib.stride_tricks.sliding_window_view(i, window, axis=0) for i in inputs]

    return func1(*arrs, out, window, func2, *args)



from functools import wraps

import numpy as np
from bottleneck import *


def bn_move_window(func):
    """bottleneck在数据量少返回nan"""

    @wraps(func)
    def decorated(a, window, *args, axis=-1, **kwargs):
        if a.shape[axis] < window:
            return np.full_like(a, fill_value=np.nan)
        return func(a, window, *args, axis=axis, **kwargs)

    return decorated


move_argmax = bn_move_window(move_argmax)
move_argmin = bn_move_window(move_argmin)
move_max = bn_move_window(move_max)
move_mean = bn_move_window(move_mean)
move_median = bn_move_window(move_median)
move_min = bn_move_window(move_min)
move_rank = bn_move_window(move_rank)
move_std = bn_move_window(move_std)
move_sum = bn_move_window(move_sum)
move_var = bn_move_window(move_var)



"""
对指标的算子化包装
1. 包装成只支持 长表 输入和输出
2. 简化参数输入，命名参数也可当成位置参数输入
3. 通过dropna的方法，自动跳过停牌

!!!函数太多，又想要智能提示，只能手工按需补充
"""
import ta_cn.talib as ta
from ..alpha import CUMPROD
from ..alpha import FILTER
from ..alpha import LessThan
from ..alpha import RANK
from ..alpha import TS_RANK
from ..alpha import scale
from ..alpha import signedpower
from ..ema import SMA
from ..logical import IF
from ..maths import ABS
from ..maths import LN
from ..maths import MAX
from ..maths import MIN
from ..maths import SGN
from ..preprocess import demean
from ..reference import COUNT
from ..reference import DIFF
from ..reference import HHV
from ..reference import HHVBARS
from ..reference import LLV
from ..reference import LLVBARS
from ..reference import MA
from ..reference import PRODUCT
from ..reference import REF
from ..reference import SUM
from ..reference import SUMIF
from ..regress import REGRESI
from ..regress import SLOPE_YX_NB
from ..statistics import COVAR
from ..statistics import STDP
from ..utils import to_pd
from ..utils_long import dataframe_groupby_apply, series_groupby_apply

_ta1d = ta.init(mode=1, skipna=False, to_globals=False)

# 按股票分组，计算时序指标。注意，组内时序需要已经排序
BY_ASSET = 'asset'
# 按时间分组。计算横截面
BY_DATE = 'date'
# 横截面上进行行业中性化
BY_GROUP = ['date', 'group']

## TALIB, 多输入
CORREL = dataframe_groupby_apply(_ta1d.CORREL, by=BY_ASSET)
ATR = dataframe_groupby_apply(_ta1d.ATR, by=BY_ASSET, to_df=[0, 1, 2], to_kwargs={3: 'timeperiod'})

## TALIB, 单输入
LINEARREG_SLOPE = series_groupby_apply(_ta1d.LINEARREG_SLOPE, by=BY_ASSET)
SMA_TA = series_groupby_apply(_ta1d.SMA, by=BY_ASSET)
WMA = series_groupby_apply(_ta1d.WMA, by=BY_ASSET)

#
CUMPROD = series_groupby_apply(CUMPROD, by=BY_ASSET, to_kwargs={})
FILTER = dataframe_groupby_apply(FILTER, by=BY_ASSET, to_kwargs={}, dropna=False)
RANK = series_groupby_apply(RANK, by=BY_DATE, to_kwargs={})
TS_RANK = series_groupby_apply(TS_RANK, by=BY_ASSET)
LessThan = to_pd(LessThan)
scale = series_groupby_apply(scale, by=BY_DATE, to_kwargs={1: 'a'}, dropna=False)
signedpower = to_pd(signedpower)

#
IF = to_pd(IF)
ABS = to_pd(ABS)
LN = to_pd(LN)
MAX2 = to_pd(MAX)
MIN2 = to_pd(MIN)
SGN = to_pd(SGN)

#
SMA = series_groupby_apply(SMA, by=BY_ASSET, to_kwargs={1: 'timeperiod', 2: 'M'})

#
COUNT = series_groupby_apply(COUNT, by=BY_ASSET)
DIFF = series_groupby_apply(DIFF, by=BY_ASSET)
HHV = series_groupby_apply(HHV, by=BY_ASSET)
HHVBARS = series_groupby_apply(HHVBARS, by=BY_ASSET)
LLV = series_groupby_apply(LLV, by=BY_ASSET)
LLVBARS = series_groupby_apply(LLVBARS, by=BY_ASSET)
MA = series_groupby_apply(MA, by=BY_ASSET)
PRODUCT = series_groupby_apply(PRODUCT, by=BY_ASSET, dropna=False)
REF = series_groupby_apply(REF, by=BY_ASSET)
SUM = series_groupby_apply(SUM, by=BY_ASSET)
SUMIF = dataframe_groupby_apply(SUMIF, by=BY_ASSET)

SLOPE_YX_NB = dataframe_groupby_apply(SLOPE_YX_NB, by=BY_ASSET)
REGRESI4 = dataframe_groupby_apply(REGRESI, by=BY_ASSET, to_df=[0, 1, 2, 3], to_kwargs={4: 'timeperiod'})

COVAR = dataframe_groupby_apply(COVAR, by=BY_ASSET)
STDP = series_groupby_apply(STDP, by=BY_ASSET)

# 行业中性。demean法
indneutralize = dataframe_groupby_apply(demean, by=BY_GROUP, to_df=[0, 'group'], to_kwargs={})



"""
文件分组处理相关
"""
import os
import pathlib
from typing import Union, List

import pandas as pd
from loguru import logger


def load_parquet(path: Union[str, pathlib.Path], pattern: Union[str, List[str]]):
    """加载多个parquet文件

    Parameters
    ----------
    path: str
        目录
    pattern: str or list
        模式。如果是列表，将依次加载合并

    Returns
    -------
    pd.DataFrame

    Examples
    --------
    >>> load_parquet(path, f'{year}*')

    """
    if isinstance(pattern, str):
        # 一个过滤条件时，拼接一组
        return pd.concat([pd.read_parquet(_) for _ in pathlib.Path(path).glob(pattern)])
    if isinstance(pattern, list):
        # 多个过滤条件时，拼接多组
        return pd.concat([load_parquet(path, _) for _ in pattern])


def merge_parquet(left, right, how='left', **merge_kwargs):
    """

    Parameters
    ----------
    left: tuple
        path, pattern
    right: tuple
        path, pattern
    how: str
        合并方案
    merge_kwargs:
        pd.merge的参数

    Returns
    -------

    """
    l_ = load_parquet(*left)
    r_ = load_parquet(*right)
    return pd.merge(l_, r_, how=how, **merge_kwargs)


def _dataframe_make_key(df: pd.DataFrame, key='asset'):
    """创建股票分组，此为参考示例，实际使用需重新实现
    """
    # 过滤停牌。计算技术指标和和横截面时会剔除停牌，但计算板块和指数时，停牌也参与计算
    # JoinQuant中行情默认停牌也有记录，标记成停牌
    # TuShare没有停牌日的行情，只能通过unstack还原数据后再fillna
    df = df[df['paused'] == 0]  # .copy()
    df = df[~df[key].str.endswith('.BJ')]  # .copy()
    # [5: 6]一位用来当关键字，数据量大就用两位[4:6]
    df['key'] = df[key].str[5:6]
    return df


def split_by_symbol(df: pd.DataFrame, date: str, output: Union[str, pathlib.Path],
                    set_index_keys=['asset', 'date'], group_by=['key'], drop_columns=['key']):
    """按股票代码生成的关键字进行时序方向分割

    1. 直接按股票代码直接分割会因为产生的文件太多，反而处理慢
    2. 复合索引占空间，多一些字段才划算

    Parameters
    ----------
    df: pd.DataFrame
    date: str
        划分的起始日期
    output:str
        输出路径
    set_index_keys: list
        设置成索引的列。先合约代码后时间
    group_by: list
        分组列
    drop_columns: list
        放弃列

    Examples
    --------
    >>> split_by_symbol(df, year, output, set_index_keys=['code', 'time'], group_by=['key'], drop_columns=['key'])
    >>> split_by_symbol(df, year, output, set_index_keys=['ts_code', 'trade_date'], group_by=['key'], drop_columns=['key', 'change', 'pct_chg'])

    """
    output = pathlib.Path(output)
    output.mkdir(parents=True, exist_ok=True)

    df = df.set_index(set_index_keys).sort_index()
    grouped = df.groupby(by=group_by)
    for key, group in grouped:
        logger.info('pid:{}, key:{}, date:{}', os.getpid(), key, date)
        # 删除没必要保存的字段
        group.drop(columns=drop_columns, inplace=True)
        group.to_parquet(output / f'{key}__{date}.parquet', compression='gzip')


def split_by_column(df: pd.DataFrame, date: str, output: Union[str, pathlib.Path],
                    set_index_keys=['date', 'asset'], drop_columns=['change', 'pct_chg']):
    """按字段进行分割

    1. 二维，一个字段一个文件
    2. 如果上市有先后，会导致大量的空白区域，占空间
    3. 停牌导致数据中断，部分计算指标函数会错误

    Parameters
    ----------
    df: pd.DataFrame
    date: str
        划分的起始日期
    output: str
        输出路径
    set_index_keys: list
        设置成索引的列。先时间后合约代码
    drop_columns: list
        放弃列

    """
    # 设置复合索引
    output = pathlib.Path(output)
    output.mkdir(parents=True, exist_ok=True)

    df = df.set_index(set_index_keys).sort_index()
    for col in df.columns:
        # 跳过没有必要保存的字段
        if col in drop_columns:
            continue
        logger.info('pid:{}, col:{}, date:{}', os.getpid(), col, date)
        # 转换保存
        d = df[col].unstack()
        d.to_parquet(output / f'{col}__{date}.parquet', compression='gzip')


def dataframe_calc(df: pd.DataFrame, groupbys: List[str], funcs: List):
    """dataframe计算模板

    Parameters
    ----------
    df: pd.DataFrame
        复合索引的数据
    groupbys: list
        处理方向，None表示不用groupby,直接处理
    funcs:
        不分组，直接处理。如一列与另一列的处理
        分组，组内进行计算，比如算指标，算排序等

    Returns
    -------
    pd.DataFrame

    Examples
    --------
    >>> df = dataframe_calc(df, [None, 'asset', 'date'], [func_all, func_ts, func_cs])

    """
    for g, f in zip(groupbys, funcs):
        if g is None:
            # 不分组
            df = f(df)
        else:
            # 注意分组方向
            df = df.groupby(by=g).apply(f)

    return df



"""
宽表处理工具
"""
from functools import wraps

import numpy as np

from ta_cn.utils import pd_to_np


def get_uint_type(i):
    """通过数据大小得到合适的类型，用于减少内存"""
    # np.uint8, np.uint128, np.uint256 太小或太大都没什么意义
    int_types = [np.uint16, np.uint32, np.uint64]
    # np.iinfo(it).max
    # 65535
    # 4294967295
    # 18446744073709551615
    if i <= 65535:
        return np.uint16
    if i <= 4294967295:
        return np.uint32
    else:
        return np.uint64


def pushna(arr, direction='down'):
    """将非空数据按方向移动

    Parameters
    ----------
    arr: np.darray
    direction: str
        up, down, left, right

    Returns
    -------
    arr
        重排后的数据
    row
        行索引。将用于还原
    col
        列索引。将用于还原

    References
    ----------
    https://stackoverflow.com/questions/32062157/move-non-empty-cells-to-the-left-in-pandas-dataframe
    https://stackoverflow.com/questions/39361839/move-non-empty-cells-to-left-in-grouped-columns-pandas

    """
    if direction is None:
        return arr, None, None

    if direction == 'down':
        row = (~np.isnan(arr)).argsort(axis=0, kind='stable').astype(get_uint_type(arr.shape[0]))
        col = np.arange(arr.shape[1])[None]
    if direction == 'up':
        row = np.isnan(arr).argsort(axis=0, kind='stable').astype(get_uint_type(arr.shape[0]))
        col = np.arange(arr.shape[1])[None]
    if direction == 'left':
        col = np.isnan(arr).argsort(axis=1, kind='stable').astype(get_uint_type(arr.shape[1]))
        row = np.arange(arr.shape[0])[:, None]
    if direction == 'right':
        col = (~np.isnan(arr)).argsort(axis=1, kind='stable').astype(get_uint_type(arr.shape[1]))
        row = np.arange(arr.shape[0])[:, None]

    return arr[row, col], row, col


def pullna(arr, row, col):
    """pushna的还原

    Parameters
    ----------
    arr: np.array
    row
        行索引
    col
        列索引

    Examples
    --------
    >>> a, row, col = pushna(df)
    >>> print(pullna(a, row, col))

    """
    if row is None:
        return arr

    # col 过于简单，可以通过direction推断出来

    tmp = np.empty_like(arr)
    tmp[row, col] = arr
    return tmp


class WArr(np.ndarray):
    """宽表计算中间体，用于进行输入的堆叠，输出的还原，不可修改"""
    arr = None  # 堆叠后的数据
    row = None
    col = None
    direction = None
    _raw = None

    def raw(self):
        if self._raw is None:
            self._raw = pullna(self.arr, self.row, self.col)
        return self._raw

    @classmethod
    def from_args(cls, arr, row, col, direction):
        c = cls(shape=(1,))
        c.arr, c.row, c.col, c.direction = arr, row, col, direction
        return c

    @classmethod
    def from_array(cls, arr, direction):
        return cls.from_args(*pushna(pd_to_np(arr), direction), direction)

    @classmethod
    def from_obj(cls, obj, direction):
        if isinstance(obj, cls):
            # 方向一样就直接返回
            if direction == obj.direction:
                return obj
            # 不一样就还原数据
            return cls.from_array(obj.raw(), direction)
        else:
            return cls.from_array(obj, direction)

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            return self.from_args(self.arr * other, self.row, self.col, self.direction)
        if isinstance(other, WArr):
            return self.from_array(self.raw() * other.raw(), self.direction)
        else:
            return self.from_array(self.raw() * other, self.direction)

    def __rmul__(self, other):
        if isinstance(other, (int, float)):
            return self.from_args(other * self.arr, self.row, self.col, self.direction)
        if isinstance(other, WArr):
            return self.from_array(other.raw() * self.raw(), self.direction)
        else:
            return self.from_array(other * self.raw(), self.direction)

    def __gt__(self, other):
        if isinstance(other, (int, float)):
            return self.from_args(self.arr > other, self.row, self.col, self.direction)
        if isinstance(other, WArr):
            return self.from_array(self.raw() > other.raw(), self.direction)
        else:
            return self.from_array(self.raw() > other, self.direction)

    def __lt__(self, other):
        if isinstance(other, (int, float)):
            return self.from_args(self.arr < other, self.row, self.col, self.direction)
        if isinstance(other, WArr):
            return self.from_array(self.raw() < other.raw(), self.direction)
        else:
            return self.from_array(self.raw() < other, self.direction)

    def __ge__(self, other):
        if isinstance(other, (int, float)):
            return self.from_args(self.arr >= other, self.row, self.col, self.direction)
        if isinstance(other, WArr):
            return self.from_array(self.raw() >= other.raw(), self.direction)
        else:
            return self.from_array(self.raw() >= other, self.direction)

    def __le__(self, other):
        if isinstance(other, (int, float)):
            return self.from_args(self.arr <= other, self.row, self.col, self.direction)
        if isinstance(other, WArr):
            return self.from_array(self.raw() <= other.raw(), self.direction)
        else:
            return self.from_array(self.raw() <= other, self.direction)

    def __eq__(self, other):
        if isinstance(other, (int, float)):
            return self.from_args(self.arr == other, self.row, self.col, self.direction)
        if isinstance(other, WArr):
            return self.from_array(self.raw() == other.raw(), self.direction)
        else:
            return self.from_array(self.raw() == other, self.direction)

    def __add__(self, other):
        if isinstance(other, (int, float)):
            return self.from_args(self.arr + other, self.row, self.col, self.direction)
        if isinstance(other, WArr):
            return self.from_array(self.raw() + other.raw(), self.direction)
        else:
            return self.from_array(self.raw() + other, self.direction)

    def __radd__(self, other):
        if isinstance(other, (int, float)):
            return self.from_args(other + self.arr, self.row, self.col, self.direction)
        if isinstance(other, WArr):
            return self.from_array(other.raw() + self.raw(), self.direction)
        else:
            return self.from_array(other + self.raw(), self.direction)

    def __sub__(self, other):
        if isinstance(other, (int, float)):
            return self.from_args(self.arr - other, self.row, self.col, self.direction)
        if isinstance(other, WArr):
            return self.from_array(self.raw() - other.raw(), self.direction)
        else:
            return self.from_array(self.raw() - other, self.direction)

    def __rsub__(self, other):
        if isinstance(other, (int, float)):
            return self.from_args(other - self.arr, self.row, self.col, self.direction)
        if isinstance(other, WArr):
            return self.from_array(other.raw() - self.raw(), self.direction)
        else:
            return self.from_array(other - self.raw(), self.direction)

    def __truediv__(self, other):
        if isinstance(other, (int, float)):
            return self.from_args(self.arr / other, self.row, self.col, self.direction)
        if isinstance(other, WArr):
            return self.from_array(self.raw() / other.raw(), self.direction)
        else:
            return self.from_array(self.raw() / other, self.direction)

    def __rtruediv__(self, other):
        if isinstance(other, (int, float)):
            return self.from_args(other / self.arr, self.row, self.col, self.direction)
        if isinstance(other, WArr):
            return self.from_array(other.raw() / self.raw(), self.direction)
        else:
            return self.from_array(other / self.raw(), self.direction)

    def __pow__(self, power, modulo=None):
        if isinstance(power, (int, float)):
            return self.from_args(self.arr ** power, self.row, self.col, self.direction)
        if isinstance(power, WArr):
            return self.from_array(self.raw() ** power.raw(), self.direction)
        else:
            return self.from_array(self.raw() ** power, self.direction)

    def __rpow__(self, other):
        if isinstance(other, (int, float)):
            return self.from_args(other ** self.arr, self.row, self.col, self.direction)
        if isinstance(other, WArr):
            return self.from_array(other.raw() ** self.raw(), self.direction)
        else:
            return self.from_array(other ** self.raw(), self.direction)

    def __neg__(self):
        return self.from_args(-self.arr, self.row, self.col, self.direction)

    def __and__(self, other):
        if isinstance(other, WArr):
            return self.from_array(self.raw() & other.raw(), self.direction)
        else:
            return self.from_array(self.raw() & other, self.direction)

    def __or__(self, other):
        if isinstance(other, WArr):
            return self.from_array(self.raw() | other.raw(), self.direction)
        else:
            return self.from_array(self.raw() | other, self.direction)


def get_raw_arr(arr):
    if isinstance(arr, WArr):
        return arr.raw()
    return arr


def wide_wraps(func, direction='down', input_num=1, output_num=1, to_kwargs={1: 'timeperiod'}):
    """将二维函数包装成可堆叠函数

    Parameters
    ----------
    func:
        1. 必须是二维数据处理函数
        2. 由于数据已经堆叠。内部处理时最好能跳过空行或空列
    direction
        数据堆叠方向。时序堆叠到最后。
    input_num
    output_num

    """

    @wraps(func)
    def decorated(*args, **kwargs):
        _kwargs = {k: args[i] for i, k in to_kwargs.items() if i < len(args)}
        _args = [WArr.from_obj(v, direction) for i, v in enumerate(args) if i not in to_kwargs]
        arg0 = _args[0]

        # 输入+计算
        if input_num == 1:
            outputs = func(arg0.arr, **kwargs, **_kwargs)
        else:
            outputs = func(*[v.arr for v in _args], **kwargs, **_kwargs)

        # 输出
        if output_num == 1:
            return WArr.from_args(outputs, arg0.row, arg0.col, direction)
        else:
            return [WArr(v, arg0.row, arg0.col, direction) for v in outputs]

    return decorated



"""
通达信公式转alpha101

101 Formulaic Alphas
文档质量一般
1. 部分函数即有全小写，又有大小写混合
2. 很多参数应当是整数，但输入是小数，不得不做修正
"""
from ta_cn.imports import long as L
from ta_cn.utils import round_a_i, round_a_a_i

correlation = round_a_a_i(L.CORREL)
decay_linear = round_a_i(L.WMA)

LessThan = L.LessThan
rank = L.RANK
ts_rank = round_a_i(L.TS_RANK)
scale = L.scale
SignedPower = L.signedpower

IF = L.IF
abs = L.ABS
log = L.LN  # 这里是用的自然对数
MAX = L.MAX2
MIN = L.MIN2
sign = L.SGN

delta = round_a_i(L.DIFF)
ts_max = round_a_i(L.HHV)
ts_argmax = round_a_i(L.HHVBARS)
ts_min = round_a_i(L.LLV)
ts_argmin = round_a_i(L.LLVBARS)
product = round_a_i(L.PRODUCT)
delay = round_a_i(L.REF)
sum = round_a_i(L.SUM)

covariance = round_a_a_i(L.COVAR)
stddev = round_a_i(L.STDP)  # 引入的是全体标准差

indneutralize = L.indneutralize


# 部分别名，这样官方公式可以减少改动
Ts_Rank = ts_rank
IndNeutralize = indneutralize
Ts_ArgMax = ts_argmax
Ts_ArgMin = ts_argmin
LessThan = LessThan
min = MIN
max = MAX
Sign = sign
Log = log



import numpy as _np

from ta_cn.utils import np_to_pd


def _MA_EMA(real, timeperiod, com=None, span=None, alpha=None):
    """内部函数，已经废弃。先计算MA做第一个值，然后再算EMA
    二维矩阵如果开头有NaN，那么求均值的位置就不统一，得做特别处理
    好不容易想出来的，删了可惜
    """
    real = np_to_pd(real, copy=True)  # 开头部分将写入SMA
    ma = np_to_pd(_np.zeros_like(real), copy=False)  # 来计算sma

    # 取最长位置, 用于计算SMA，没有必要全算一次MA
    max_end = real.notna().idxmax() + timeperiod
    if hasattr(max_end, 'max'):
        max_end = max_end.max()

    # 计算ma
    ma[:max_end] = real[:max_end].rolling(window=timeperiod, min_periods=timeperiod).mean()
    # 计算需要复制的区域,此区前部分为NaN,最后为mean
    mask = ma.isna().shift(fill_value=True)
    real[mask] = ma
    return real.ewm(com=com, span=span, alpha=alpha, min_periods=0, adjust=False).mean()


"""

"""
import numpy as _np

import ta_cn.bn_wraps as _bn


def signedpower(real, n):
    """保持符号的次方"""
    with _np.errstate(invalid='ignore', divide='ignore'):
        return _np.sign(real) * (abs(real) ** n)


def TS_RANK(real, timeperiod: int):
    """滚动rank

    real, timeperiod: int = 10
    """
    t1 = _bn.move_rank(real, window=timeperiod, axis=0)
    return (t1 + 1.) / 2.


def RANK(real):
    """横截面rank"""
    pct: bool = True

    if real.ndim == 2:
        t1 = _bn.nanrankdata(real, axis=1)
        t2 = _np.nansum(~_np.isnan(real), axis=1, keepdims=True)
    else:
        t1 = _bn.nanrankdata(real)
        t2 = _np.nansum(~_np.isnan(real))
    if pct:
        return t1 / t2
    else:
        return t1


def scale(real, a):
    """横截面缩放

    scale(real, a=1)
    """
    if real.ndim == 2:
        b = _np.nansum(abs(real), axis=1, keepdims=True)
    else:
        b = _np.nansum(abs(real), keepdims=True)

    return real / b * a


def LessThan(x, y):
    # ValueError: Can only compare identically-labeled Series objects
    return x - y < 0


def split_adjust(real):
    """将后复权因子转成前复权"""
    if isinstance(real, _np.ndarray):
        return real / real[-1]
    else:
        return real / real.iloc[-1]


def FILTER(A, condition):
    return _np.where(condition, A, 0)


def CUMPROD(A):
    if A.ndim == 2:
        return _np.cumprod(A, axis=0)
    else:
        return _np.cumprod(A)



"""宽表不好处理的算子，转成长表进行处理再还原"""
from ta_cn.utils import np_to_pd
from ta_cn.utils_wide import get_raw_arr, WArr
from .long import indneutralize as _indneutralize


def indneutralize(x, group):
    """行业中性化

    对宽表处理实在过于麻烦，所以将其转换成了长表
    """

    x1 = np_to_pd(get_raw_arr(x)).stack(dropna=False)
    g1 = np_to_pd(get_raw_arr(group)).stack(dropna=False)

    x1.index.names = ['date', 'asset']
    g1.index.names = ['date', 'asset']

    r = _indneutralize(x1, g1)

    return WArr.from_array(r.unstack(), direction=None)



"""
长表处理工具
"""
from functools import wraps

import pandas as pd

from ta_cn.utils import to_pd


def series_groupby_apply(func, by='asset', to_kwargs={1: 'timeperiod'}, dropna=True):
    """普通指标转换成按分组处理的指标。只支持单参数

    Parameters
    ----------
    func
    by
    to_kwargs

    Notes
    -----
    只能处理Series

    """

    @wraps(func)
    def decorated(*args, **kwargs):
        # 参数位置调整，实现命命参数简化
        _kwargs = {k: args[i] for i, k in to_kwargs.items() if i < len(args)}
        s1 = args[0]

        if dropna:
            s2 = s1.dropna()
        else:
            s2 = s1

        if len(s2) == 0:
            return pd.Series(index=s1.index, dtype=float)

        s3 = s2.groupby(by=by, group_keys=False).apply(to_pd(func), **_kwargs, **kwargs)

        if len(s1) == len(s2):
            return s3
        else:
            # 由于长度变化了，只能重整长度
            return pd.Series(s3, index=s1.index)

    return decorated


def dataframe_groupby_apply(func, by='asset', to_kwargs={2: 'timeperiod'}, to_df=[0, 1], dropna=True):
    """普通指标转换成按分组处理的指标。支持多输入

    Parameters
    ----------
    func
    by
    dropna
    to_df
    to_kwargs

    Notes
    -----
    即能处理DataFrame，又能处理Series,但考虑到效率，单输入时使用series_groupby_apply

    """

    def get(i, k, args, kwargs):
        if i == k:
            return args[i]
        if isinstance(k, str):
            v = kwargs.get(k, None)
            if v is None:
                return args[i]
            return v

    @wraps(func)
    def decorated(*args, **kwargs):
        _kwargs = {k: args[i] for i, k in to_kwargs.items() if i < len(args)}
        s1 = pd.DataFrame({k: get(i, k, args, kwargs) for i, k in enumerate(to_df)})
        if dropna:
            s2 = s1.dropna()
        else:
            s2 = s1
        if len(s2) == 0:
            return pd.Series(index=s1.index, dtype=float)

        s3 = s2.groupby(by=by, group_keys=False).apply(to_pd(dataframe_split(func)), **_kwargs)

        if len(s1) == len(s2):
            return s3
        else:
            # 由于长度变化了，只能重整长度
            return pd.Series(s3, index=s1.index)

    return decorated


def dataframe_split(func):
    """将第一个DataFrame分拆传到指定函数"""

    @wraps(func)
    def decorated(df: pd.DataFrame, *args, **kwargs):
        ss = df.to_dict(orient='series')
        args_input = [v for k, v in ss.items() if isinstance(k, int)]
        kwargs_input = {k: v for k, v in ss.items() if not isinstance(k, int)}
        return func(*args_input, *args, **kwargs_input, **kwargs)

    return decorated



import warnings

import numba
import numpy as _np
import pandas as _pd

import ta_cn.bn_wraps as _bn
from ta_cn import talib as ta
from .nb import numpy_rolling_apply, _rolling_func_1_nb, _rolling_func_2_nb
from .utils import pd_to_np

_ta1d = ta.init(mode=1, skipna=False)
_ta2d = ta.init(mode=2, skipna=False)


@numba.jit(nopython=True, cache=True, nogil=True)
def _slope_nb(y, x, m_x):
    """slope线性回归斜率。由于x固定，所以提前在外部计算，加快速度"""
    m_y = _np.mean(y)
    return _np.sum((x - m_x) * (y - m_y)) / _np.sum((x - m_x) ** 2)


@numba.jit(nopython=True, cache=True, nogil=True)
def _slope_yx_nb(y, x):
    """slope线性回归斜率。y与x是一直变化的"""
    m_x = _np.mean(x)
    m_y = _np.mean(y)
    return _np.sum((x - m_x) * (y - m_y)) / _np.sum((x - m_x) ** 2)


def FORCAST(real, timeperiod):
    """线性回归预测值

    FORCAST(real, timeperiod=14)

    Parameters
    ----------
    real
    timeperiod

    References
    ----------
    https://github.com/TA-Lib/ta-lib/blob/master/src/ta_func/ta_LINEARREG.c

    """
    return _ta2d.LINEARREG(real, timeperiod=timeperiod)


def SLOPE(real, timeperiod):
    """线性回归斜率

    SLOPE(real, timeperiod=14)

    Parameters
    ----------
    real
    timeperiod

    References
    ----------
    https://github.com/TA-Lib/ta-lib/blob/master/src/ta_func/ta_LINEARREG_SLOPE.c

    """
    return _ta2d.LINEARREG_SLOPE(real, timeperiod=timeperiod)


def SLOPE_NB(real, timeperiod):
    """numba版，输出结果与LINEARREG_SLOPE一样

    SLOPE_NB(real, timeperiod=14)
    """
    x = _np.arange(timeperiod)
    m_x = _np.mean(x)
    return numpy_rolling_apply([pd_to_np(real)], timeperiod, _rolling_func_1_nb, _slope_nb, x, m_x)


def SLOPE_YX_NB(real0, real1, timeperiod):
    """与talib.BETA不一样，talib将price转return后再回归

    SLOPE_YX_NB(real0, real1, timeperiod=30)
    """
    return numpy_rolling_apply([pd_to_np(real0), pd_to_np(real1)],
                               timeperiod, _rolling_func_2_nb, _slope_yx_nb)


def ts_simple_regress(y, x, window=10):
    """滚动一元线性回归。

    由于利用了bottleneck的滚动功能，比SLOPE_YX_NB快一些，但精度有少量差异，后期需再验证

    Parameters
    ----------
    y: array_like
        回归因变量
    x: array_like
        回归自变量
    window
        移动计算窗口长度

    Returns
    -------
    intercept_hat: ndarray
        回归截距项
    beta_hat: ndarray
        回归系数项
    residual_hat: ndarray
        回归残差项

    """
    xy_ts_sum = _bn.move_sum(_np.multiply(x, y), window=window, axis=0)
    xx_ts_sum = _bn.move_sum(_np.multiply(x, x), window=window, axis=0)
    x_bar = _bn.move_mean(x, window=window, axis=0)
    y_bar = _bn.move_mean(y, window=window, axis=0)

    up = xy_ts_sum - _np.multiply(x_bar, y_bar) * window
    down = xx_ts_sum - _np.multiply(x_bar, x_bar) * window
    beta_hat = up / down
    intercept_hat = y_bar - _np.multiply(beta_hat, x_bar)
    residual_hat = y - intercept_hat - _np.multiply(beta_hat, x)

    return intercept_hat, beta_hat, residual_hat


@numba.jit(nopython=True, cache=True, nogil=True)
def _ts_ols_nb(y, x):
    """使用可逆矩阵计算多元回归。

    由于sliding_window_view后的形状再enumerate后比较特殊，所以原公式的转置进行了调整
    """
    return _np.dot((_np.dot(_np.linalg.inv(_np.dot(x, x.T)), x)), y)


@numba.jit(nopython=True, cache=True, nogil=True)
def _cs_ols_nb(y, x):
    """使用可逆矩阵计算多元回归。

    标准的多元回归
    """
    return _np.dot((_np.dot(_np.linalg.inv(_np.dot(x.T, x)), x.T)), y)


def yx_rolling_apply(y, x, window, func1, func2, *args):
    """滚动应用方法。目前用于计算时序上的滚动回归"""
    out = _np.empty_like(x)
    try:
        # 可能出现类似int无法设置nan的情况
        out[:window] = _np.nan
    except:
        out[:window] = 0

    arrs = [_np.lib.stride_tricks.sliding_window_view(i, window, axis=0) for i in (y, x)]

    return func1(*arrs, out, window, func2, *args)


warnings.filterwarnings("ignore", category=numba.NumbaPerformanceWarning)


# NumbaPerformanceWarning: np.dot() is faster on contiguous arrays, called on (array(float64, 2d, C), readonly array(float64, 1d, A))
#   out[i + timeperiod - 1] = func(yy, xx, *args)
@numba.jit(nopython=True, cache=True, nogil=True)
def _rolling_func_yx_nb(y, x, out, timeperiod, func, *args):
    """滚动多元"""
    if x.ndim == 3:
        for i, (yy, xx) in enumerate(zip(y, x)):
            out[i + timeperiod - 1] = func(yy, xx, *args)

    return out


def ts_multiple_regress(y, x, timeperiod=10, add_constant=True):
    """时序上滚动多元线性回归

    Parameters
    ----------
    y: 1d array
        因变量。一维
    x: 2d array
        自变量。二维。一列为一个特征
    timeperiod:int
        周期
    add_constant: bool
        是否添加常量

    Returns
    -------
    coef:
        系数。与x形状类似，每个特性占一例。时序变化，所以每天都有一行
    residual:
        残差。与y形状类似，由实际y-预测y而得到

    """
    _y = pd_to_np(y)
    _x = pd_to_np(x)
    if add_constant:
        tmp = _np.ones(shape=(_x.shape[0], _x.shape[1] + 1))
        tmp[:, 1:] = _x
        _x = tmp
    coef = yx_rolling_apply(_y, _x, timeperiod, _rolling_func_yx_nb, _ts_ols_nb)
    residual = _y - _np.sum(_x * coef, axis=1)
    return coef, residual


def multiple_regress(y, x, add_constant=True):
    """横截面上的多元回归。主要用于中性化多元回归场景

    需要先按日期进行groupby，然后再应用回归函数
    """
    _y = pd_to_np(y)
    _x = pd_to_np(x)
    if add_constant:
        tmp = _np.ones(shape=(_x.shape[0], _x.shape[1] + 1))
        tmp[:, 1:] = _x
        _x = tmp
    coef = _cs_ols_nb(_y, _x)
    residual = _y - _np.sum(_x * coef, axis=1)
    return coef, residual


def REGRESI(y, *args, timeperiod=60):
    if isinstance(y, _pd.Series):
        x = _pd.concat(args, axis=1)
    else:
        x = _np.concatenate(args, axis=1)
    coef, resi = ts_multiple_regress(y, x, timeperiod=timeperiod, add_constant=True)
    return resi



"""
导入公式清单

公式分布在多个文件，分别导入工作量大，所以直接导入本文件即可

以后遇到支持开拓者、文华一类的公式语言时，只要import as 做好映射即可
"""
from .ema import *
from .logical import *
from .maths import *
from .over_bought_over_sold import *
from .pressure_support import *
from .reference import *
from .statistics import *
from .trend import *
from .volume import *

# from .slow import * # 计算慢的公式，可以找到替代

# 以下只是为了IDE代码格式化时不删除import
try:
    KDJ
    BOLL
    VR
    VARP
    TRIX
    BETWEEN
    ROUND
    DMA
    LLVBARS
    # CCI
except:
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author     :wukan
# @License    :(C) Copyright 2022, wukan
# @Date       :2022-06-16
"""
Examples
--------
只要替换导入即可直接支持二维矩阵
>>> import talib as ta
>>> import ta_cn.talib as ta

"""
from functools import wraps, reduce

import numpy as np
import talib as _talib
from talib import abstract as _abstract


def tafunc_nditer_1(tafunc, args, kwargs, input_names, output_names, skipna):
    """直接调用talib"""

    def ALL_NOTNA(*args):
        """多输入，同位置没有出现过nan,标记成True"""
        return reduce(lambda x, y: np.logical_and(x, ~np.isnan(y)), [True] + list(args))

    # 不跳过空值，直接调用函数
    if not skipna:
        return tafunc(*args, **kwargs)

    # 取一个非数字，得用于得到形状
    real = args[0]

    output_num = len(output_names)

    outputs = [np.full_like(real, fill_value=np.nan) for _ in output_names]

    _notna = ALL_NOTNA(*args)
    # 只有不连续的nan才需要做切片 https://www.cnpython.com/qa/352363
    # TODO: https://stackoverflow.com/questions/41721674/find-consecutive-repeated-nan-in-a-numpy-array/41722059#41722059
    _in = [v[_notna] for v in args]

    if len(_in[0]) > 0:
        ta_out = tafunc(*_in, **kwargs)
        if output_num == 1:
            ta_out = [ta_out]

        for _i, _o in zip(outputs, ta_out):
            _i[_notna] = _o

    # 输出
    return outputs[0] if output_num == 1 else tuple(outputs)


def tafunc_nditer_2(tafunc, args, kwargs, input_names, output_names,
                    skipna, order='F'):
    """内部按列迭代函数，支持timeperiod等命名参数向量化

    Parameters
    ----------
    tafunc
        计算单列的函数
    args
        位置参数
    kwargs
        命名参数
    output_names: list
        tafunc输出参数名
    skipna:
        如想跳过空值，需将数据堆叠到首或尾，实现连续计算
    order:
        F, 按列进行遍历
        C, 按行进行遍历

    Returns
    -------
    tuple
        输出元组

    """

    def num_to_np(x, like):
        """将单字数字转成矩阵"""
        if hasattr(x, "__getitem__"):
            # 长度不足时，用最后一个值填充之后的参数
            y = np.full_like(like, fill_value=x[-1])
            # 长度超出时，截断
            y[:len(x)] = x[:len(y)]
            return y
        # 单一值，填充成唯一值
        return np.full_like(like, fill_value=x)

    def last_isna(x):
        # 只检查最后一行
        return np.any([y[-1] != y[-1] for y in x])

    real = args[0]

    if real.ndim == 1:
        return tafunc_nditer_1(tafunc, args, kwargs, input_names, output_names, skipna)

    # =====以下是二维======
    inputs = [*args]

    # 输出缓存
    outputs = [np.full_like(real, fill_value=np.nan) for _ in output_names]
    kwargs = {k: num_to_np(v, real[0]) for k, v in kwargs.items()}

    # 只有一行输入时需要特别处理
    with np.nditer(inputs + outputs,
                   flags=['external_loop'] if real.shape[0] > 1 else None,
                   order=order,
                   op_flags=[['readonly']] * len(inputs) + [['writeonly']] * len(outputs)) as it:
        for i, in_out in enumerate(it):
            if real.shape[0] == 1:
                # 需要将0维array改成1维，否则talib报错
                in_out = [v.reshape(1) for v in in_out]

            _in = in_out[:len(inputs)]  # 分离输入
            # 最后一行出现了空
            if last_isna(_in):
                continue
            _out = in_out[-len(outputs):]  # 分离输出

            # 切片得到每列的参数
            _kw = {k: v[i] for k, v in kwargs.items()}

            # 计算并封装
            ta_out = tafunc(*_in, **_kw)
            if not isinstance(ta_out, tuple):
                ta_out = tuple([ta_out])

            for _i, _o in zip(_out, ta_out):
                _i[...] = _o

    # 输出
    if len(outputs) == 1:
        return outputs[0]
    return outputs


def ta_decorator(func, mode, input_names, output_names, skipnan):
    # 设置对应处理函数
    ff = {1: tafunc_nditer_1, 2: tafunc_nditer_2}.get(mode)

    @wraps(func)
    def decorated(*args, **kwargs):
        return ff(func, args, kwargs, input_names, output_names, skipnan)

    return decorated


def init(mode=1, skipna=False, to_globals=False):
    """初始化环境

    Parameters
    ----------
    mode: int
        1: 输入数据支持一维矩阵。数据使用位置，周期等使用命名
        2. 输入参数支持一维向量。数据使用位置，周期等使用命名。否则报错
    skipna: bool
        是否跳过空值。跳过空值功能会导致计算变慢。
        - 确信数据不会中途出现空值建议设置成False, 加快计算
    to_globals: bool
        注册到包中

    """
    assert mode in (1, 2)

    # print(f'ta_cn mode: {mode}, skipna: {skipna}')
    #
    # if mode == 1:
    #     print(f'\t1. 输入一维数据，支持skipna跳过空值。必须使用命名参数传入周期，使用位置参数传入数据。')
    # if mode == 2:
    #     print(f'\t2. 输入二维数据，支持skipna跳过空值。必须使用命名参数传入周期，使用位置参数传入数据。周期参数由只支持标量升级为一维向量')

    class TA_CN_LIB:
        pass

    lib = TA_CN_LIB()
    for i, func_name in enumerate(_talib.get_functions()):
        """talib遍历"""
        _ta_func = getattr(_talib, func_name)
        info = _abstract.Function(func_name)._Function__info
        output_names = info['output_names']
        input_names = info['input_names']

        # 创建函数
        f = ta_decorator(_ta_func, mode, input_names, output_names, skipna)
        setattr(lib, func_name, f)
        if to_globals:
            globals()[func_name] = f

    return lib


# =============================================

TA_COMPATIBILITY_DEFAULT = 0  # 使用MA做第一个值
TA_COMPATIBILITY_METASTOCK = 1  # 使用Price做第一个值

_COMPATIBILITY_ENABLE_ = False


def set_compatibility_enable(enable):
    """talib兼容性设置"""
    global _COMPATIBILITY_ENABLE_
    _COMPATIBILITY_ENABLE_ = enable


def set_compatibility(compatibility):
    """talib兼容性设置"""
    global _COMPATIBILITY_ENABLE_
    if _COMPATIBILITY_ENABLE_:
        print('do talib.set_compatibility', compatibility)
        _talib.set_compatibility(compatibility)



from functools import reduce

import numpy as _np

from .nb import numpy_rolling_apply, _rolling_func_1_nb, _last_nb
from .reference import SUM
from .utils import np_to_pd, num_to_np, pd_to_np


def CROSS(S1, S2):
    """向上金叉"""
    # 可能输入单个数字，需预处理
    S1 = num_to_np(S1, S2)
    S2 = num_to_np(S2, S1)

    arr = _np.zeros_like(S1, dtype=bool)
    arr[1:] = (S1 <= S2)[:-1] & (S1 > S2)[1:]
    return arr


def IF(condition, T, F):
    """序列布尔判断"""
    return _np.where(condition, T, F)


def EVERY(real, timeperiod):
    """最近timeperiod天是否全为True

    EVERY(real, timeperiod=5)
    """
    return SUM(real, timeperiod) == timeperiod


def EXIST(real, timeperiod):
    """最近timeperiod天是否存在一天为True

    EXIST(real, timeperiod=5)
    """
    return SUM(real, timeperiod) > 0


def BETWEEN(S, A, B):
    """BETWEEN(A,B,C)表示A处于B和C之间时返回1,否则返回0"""
    return ((A < S) & (S < B)) | ((A > S) & (S > B))


def VALUEWHEN(S, X):
    """条件跟随函数。当COND条件成立时,取X的当前值,否则取VALUEWHEN的上个值"""
    return np_to_pd(_np.where(S, X, _np.nan)).ffill()


def LAST(real, n, m):
    """LAST(X,A,B)，A>B，表示从前A日到前B日一致满足X条件

    LAST(real, n=20, m=10)
    """
    return numpy_rolling_apply([pd_to_np(real)], n, _rolling_func_1_nb, _last_nb, n, m)


def AND(*args):
    return reduce(lambda x, y: _np.logical_and(x, y), [True] + list(args))


def OR(*args):
    return reduce(lambda x, y: _np.logical_or(x, y), [False] + list(args))



"""
# 计算说明
计算是需要在 `快` 与 `精` 中进行平衡，大规模计算时使用`快速评估`，然后再使用工具进行高精度回测

## 核心
1. 对于单一投资品的收益率，对数收益率时序上可加
2. 对于不同投资品的收益率，应该用简单收益率，因为它在截面上有可加性

## fast，快速计算
1. 不考虑手续费、滑点等信息，否则代码过于复杂。一般用于大规模挖掘时使用，挖出的结果再进行中高精度计算
2. 使用对数收益率累加比简单收益率累乘，计算更快
3. 信号操作需要转换成持仓状态才能使用。信号一般要下移一格才是持仓
4. 做空时收益计算有误差，但在涨跌幅不大时，可以忽略
5. 收益计算，本质上是段内单利，段间复利。但以前使用的收益率累乘，相当于每天调仓为一段，变成了每天复利，
    所以如果代码是每月调仓，收益率就得换成每月，如果是不规则调仓，那就使用 (出场价/入场价-1) 进行计算
6. 持仓取值是1 0 -1还是0.5一类的呢？本质上是看操作的资金占总资金的比重
7. 如果持仓是-2 0 1 3等手数，那盈亏其实应当用diff差分，但这种情况下建议使用vectorbt

## medium, 中速计算
1. 使用快速的向量计算框架vectorbt
2. 可以灵活设置交易成本等信息

## slow, 慢速框架
1. 通常使用事件框架来实现。如backtrader、rqalpha、zipline、聚宽等
2. 与这些平台的交互都统一成因子值，此处不存因子表达式
3. 这些平台都当成算法交易执行工具进行回测

"""

from typing import Union

import numpy as np
import pandas as pd


def to_log_returns(prices: Union[pd.Series, pd.DataFrame]) -> Union[pd.Series, pd.DataFrame]:
    """对数收益率

    Parameters
    ----------
    prices:
        价格

    Returns
    -------
    对数收益率，时序上可加

    See Also
    --------
    ffn

    """
    return np.log(prices / prices.shift(1))


def to_returns(prices: Union[pd.Series, pd.DataFrame]) -> Union[pd.Series, pd.DataFrame]:
    """简单收益率

    Parameters
    ----------
    prices

    Returns
    -------
    简单收益率，横截面上可加

    See Also
    --------
    ffn

    """
    return prices / prices.shift(1) - 1


def returns_cumsum(returns: Union[pd.Series, pd.DataFrame]) -> Union[pd.Series, pd.DataFrame]:
    """对数收益率累加

    Parameters
    ----------
    returns:
        简单收益率 / 对数收益率

    Notes
    -----
    - 简单收益率: cumsum得到的结果是单利。不建议用于做空
    - 对数收益率: cumsum得到的是log(pn/p1)，本质上是复利

    """
    return returns.fillna(0.0).cumsum()


def returns_cumprod(returns: Union[pd.Series, pd.DataFrame]) -> Union[pd.Series, pd.DataFrame]:
    """简单收益率累乘

    Parameters
    ----------
    returns
        简单收益率

    Returns
    -------
    还原价格

    Notes
    -----
    - 简单收益率：cumprod得到的是复利
    - 对数收益率：不能累乘

    """
    return returns.add(1.0, fill_value=0.0).cumprod()


def returns_exp(returns: Union[pd.Series, pd.DataFrame]) -> Union[pd.Series, pd.DataFrame]:
    """累加对数收益率还原

    Parameters
    ----------
    returns
        - 简单收益率：不可exp
        - 对数收益率：不可exp
        - 累加对数收益率：可还原

    """
    return np.exp(returns)


def equal_weighted_index(prices: Union[pd.Series, pd.DataFrame]) -> pd.Series:
    """等权重指数

    将多条价格或净值等权重合成一条指数

    1. 价格得到简单收益率。只有简单收益率才能横截面加和
    2. 得到横截面平均。自动跳过NaN
    3. 简单收益率只能累乘求复利
    4. 没有做空，一直是buy and hold，可以使用 简单收益率累乘复利

    Parameters
    ----------
    prices
        价格或净值

    Returns
    -------
    pd.Series
        指数或合成净值


    """
    # 虽然是累乘，但由于只计算一列，所以速度还能接受
    return returns_cumprod(to_returns(prices).mean(axis=1))


if __name__ == '__main__':
    # 测试通过价格序列生成等权指数
    df = pd.DataFrame({
        'A': [1, 2, 3, 4, 5, 6],
        'B': [10, 20, 30, 40, 50, 60],
        'C': [100, np.nan, np.nan, 400, 500, 600],
    })
    print(equal_weighted_index(df))

    # 生成测试数据
    df = pd._testing.makeDataFrame() + 10
    df.reset_index(inplace=True, drop=True)
    df.loc[4, 'A'] = np.nan
    print(df)

    #
    log_rets = to_log_returns(df)
    log_navs = returns_cumsum(log_rets)
    navs = returns_exp(log_navs)
    print(navs)
    print(equal_weighted_index(navs))



import ta_cn.bn_wraps as _bn
import ta_cn.talib as ta
from .nb import numpy_rolling_apply, _rolling_func_1_nb, _avedev_nb, _rolling_func_2_nb, _cov_nb
from .utils import pd_to_np

_ta1d = ta.init(mode=1, skipna=False)
_ta2d = ta.init(mode=2, skipna=False)


def AVEDEV(real, timeperiod: int):
    """平均绝对偏差

    AVEDEV(real, timeperiod=20)
    """
    return numpy_rolling_apply([pd_to_np(real)], timeperiod, _rolling_func_1_nb, _avedev_nb)


def STD(real, timeperiod: int):
    """样本标准差

    STD(real, timeperiod: int = 5)
    """
    return _bn.move_std(real, window=timeperiod, axis=0, ddof=1)


def STDP(real, timeperiod: int):
    """总体标准差

    STDP(real, timeperiod: int = 5)
    """
    # TODO: 这里有问题
    # if real.ndim == 2:
    #     return _bn.move_std(real, window=timeperiod, axis=0, ddof=0)
    # else:
    #     return _ta1d.STDDEV(real, timeperiod=timeperiod)

    if real.ndim == 2:
        return _ta2d.STDDEV(real, timeperiod=timeperiod)
    else:
        return _ta1d.STDDEV(real, timeperiod=timeperiod)

    # return _bn.move_std(real, window=timeperiod, axis=0, ddof=0)
    # if real.ndim == 2:
    #     return _ta2d.STDDEV(real, timeperiod=timeperiod)
    #     return _bn.move_std(real, window=timeperiod, axis=0, ddof=0)
    # else:
    #     return _ta1d.STDDEV(real, timeperiod=timeperiod)


def VAR(real, timeperiod: int):
    """样本方差

    VAR(real, timeperiod: int = 5)
    """
    return _bn.move_var(real, window=timeperiod, axis=0, ddof=1)


def VARP(real, timeperiod: int):
    """总体方差

    VARP(real, timeperiod: int = 5)
    """
    return _bn.move_var(real, window=timeperiod, axis=0, ddof=0)


def CORREL(real0, real1, timeperiod):
    """滚动相关系数

    CORREL(real0, real1, timeperiod=30)
    """
    return _ta2d.CORREL(real0, real1, timeperiod=timeperiod)


def COVAR(real0, real1, timeperiod):
    """滚动协方差

    COVAR(real0, real1, timeperiod=30)
    """
    return numpy_rolling_apply([pd_to_np(real0), pd_to_np(real1)],
                               timeperiod, _rolling_func_2_nb, _cov_nb)



from .logical import IF
from .maths import SGN
from .reference import SUM, DIFF, REF


def OBV(real, volume, scale):  # 能量潮指标
    """能量潮指标

    OBV(real, volume, scale=1)
    """
    # 同花顺最后会除10000，但东方财富没有除 scale=1 / 10000
    return SUM(SGN(DIFF(real)) * volume, 0) * scale


def VR(close, volume, timeperiod: int):
    """VR容量比率

    VR(close, volume, timeperiod=26)
    """
    LC = REF(close, timeperiod=1)
    return SUM(IF(close > LC, volume, 0), timeperiod) / SUM(IF(close <= LC, volume, 0), timeperiod) * 100
